#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import pytest
from frocket.common.config import config, ConfigDict
from frocket.datastore.registered_datastores import get_datastore, get_blobstore


@pytest.fixture(scope="session", autouse=True)
def init_test_redis_settings():
    config['redis.host'] = os.environ.get('TEST_REDIS_HOST', config['redis.host'])
    config['redis.port'] = os.environ.get('TEST_REDIS_PORT', config['redis.port'])
    config['redis.db'] = os.environ.get('TEST_REDIS_DB', config['redis.db'])
    print(get_datastore(), get_blobstore())  # Fail on no connection, print connection details


def get_test_redis_env_variables():
    return {
        ConfigDict.to_env_variable(key): config.get(key)
        for key in ['redis.host', 'redis.port', 'redis.db', 'datastore.redis.prefix']
    }



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import time
from time import sleep
from typing import List
from abc import abstractmethod
from collections import Counter
from frocket.common.config import config
from frocket.common.metrics import MetricName
from frocket.common.tasks.base import BaseTaskRequest, TaskStatus, TaskAttemptsInfo
from frocket.common.tasks.async_tracker import AsyncJobStatusUpdater
from frocket.invoker.base_invoker import BaseInvoker, JobStatus
from frocket.invoker.jobs.job import Job

logger = logging.getLogger(__name__)


class AsyncInvoker(BaseInvoker):
    """
    Subclassing BaseInvoker to implement an async. invocation:
    first, enqueue tasks; then, poll waiting for successful completion. Handle retries according to configuration, and
    fail on timeout if necessary.
    NOTE: this class is also abstract, as it doesn't handle how tasks are actually enqueued. See concrete subclasses
    WorkQueueInvoker and AwsLambdaInvoker.
    """
    def __init__(self, job: Job):
        super().__init__(job)
        self._run_timeout_seconds = config.int("invoker.run.timeout")
        self._poll_interval_seconds = config.int("invoker.async.poll.interval.ms") / 1000
        self._log_interval_seconds = config.int("invoker.async.log.interval.ms") / 1000
        self._max_attempts = config.int("invoker.retry.max.attempts")
        self._retry_failed_interval = config.float("invoker.retry.failed.interval")
        self._retry_lost_interval = config.int("invoker.retry.lost.interval")

    def _do_run(self, task_requests: List[BaseTaskRequest],
                async_status_updater: AsyncJobStatusUpdater = None):
        """Implementing abstract BaseInvoker._do_run() as a dual-stage process: enqueue, then poll."""
        try:
            with self._metrics.measure(MetricName.ASYNC_ENQUEUE_SECONDS):
                if async_status_updater:
                    async_status_updater.update(message="Enqueuing tasks")
                self._enqueue(task_requests)
                logger.info(f"Enqueued {len(task_requests)} tasks")

            with self._metrics.measure(MetricName.ASYNC_POLL_SECONDS):
                if async_status_updater:
                    async_status_updater.update(message="Polling results")
                raw_result = self._poll(async_status_updater)
        except Exception as e:
            logger.error(str(e))
            raw_result = JobStatus(success=False, error_message=str(e), attempts_status=[])
        return raw_result

    @abstractmethod
    def _enqueue(self, requests: List[BaseTaskRequest]) -> None:
        """Actual invocation method left to subclasses"""
        pass

    def _poll(self, async_status_updater: AsyncJobStatusUpdater = None) -> JobStatus:
        """Once tasks were enqueued, the polling mechanism is shared to all subclasses: status is read from the
        datastore and analyzed."""
        total_tasks = self._job.total_tasks()
        poll_start_time = time.time()
        last_log_time = poll_start_time
        error_message = None

        while True:
            # TODO backlog to prevent the small-ish drift, sleep only the remaining time to the next 'tick'
            sleep(self._poll_interval_seconds)
            poll_running_time = time.time() - poll_start_time

            # Fetch current state - for each task index, take it latest attempt only (if more than one)
            attempts_status = self._get_attempts_status()
            latest_status_values = [at.latest_update.status for at in attempts_status]
            status_counts = Counter(latest_status_values)  # Counts instances by TaskStatus member
            parts_ended_ok = status_counts[TaskStatus.ENDED_SUCCESS]

            if async_status_updater:
                async_status_updater.update(task_counters=status_counts)

            # End polling if all done, or timed out
            if parts_ended_ok == total_tasks:
                break  # All done!
            elif poll_running_time > self._run_timeout_seconds:
                error_message = f"Query timed out after {poll_running_time:.2f} seconds"
                break

            # Periodically log progress
            should_log_progress = time.time() - last_log_time >= self._log_interval_seconds
            if should_log_progress:
                status_string = self.simple_status_string(status_counts)
                percent_done = parts_ended_ok / total_tasks * 100
                logger.info(f"{percent_done:.1f}% done in {poll_running_time:.2f}s... {status_string}")
                last_log_time = time.time()

            # Handling retries for failed and seemingly-lost tasks
            # If any failed task reached its max attempts, fail fast without trying to recover other tasks
            no_retry_failures = [at for at in attempts_status
                                 if at.latest_update.status == TaskStatus.ENDED_FAILED
                                 and at.attempt_count >= self._max_attempts]

            if len(no_retry_failures) > 0:
                error_message = f"No more retries for task(s): {[at.task_index for at in no_retry_failures]}, failing"
                break

            retries = self._build_retries(attempts_status, should_log_progress)
            if len(retries) > 0:
                self._enqueue(retries)
                logger.debug(f"Enqueued retry attempts: {retries}")

        # Poll loop has exited, return final state
        parts_failed = total_tasks - parts_ended_ok
        success = (error_message is None and parts_failed == 0)
        return JobStatus(success, error_message, attempts_status)

    def _build_retries(self,
                       attempts_status: List[TaskAttemptsInfo],
                       should_log_progress: bool) -> List[BaseTaskRequest]:
        retries = []
        failures = [at for at in attempts_status if at.latest_update.status == TaskStatus.ENDED_FAILED]

        if len(failures) > 0:
            for at in failures:
                if at.time_since_update >= self._retry_failed_interval:
                    # This already initializes a QUEUED status for the new attempt, so visible to next poll
                    retries.append(self._build_retry_task(at.task_index))
                    logger.info(f"Retrying task {at.task_index}, last failure {at.time_since_update}s ago")
                else:
                    if should_log_progress:
                        logger.info(f"Failed task {at.task_index} pending retry, failed {at.time_since_update}s ago")

        lost_attempts = [at for at in attempts_status
                         if (not at.latest_update.status.ended)
                         and at.time_since_update >= self._retry_lost_interval]

        if len(lost_attempts) > 0:
            for at in lost_attempts:
                if at.attempt_count < self._max_attempts:
                    # This already initializes a QUEUED status for the new attempt, so visible to next poll
                    retries.append(self._build_retry_task(at.task_index))
                    logger.info(f"Retrying task {at.task_index} considered lost/stuck. Last status was "
                                f"{at.latest_update.status} at {at.time_since_update} seconds ago")
                else:
                    if should_log_progress:
                        logger.info(f"No more retries for lost task {at.task_index}. May it recover before timeout!")

        return retries

    def _get_attempts_status(self) -> List[TaskAttemptsInfo]:
        # Read status of ALL attempts, and init a list of per-task attempt summaries
        tasks_status = self._datastore.tasks_status(self._request_id)

        attempts_status = [TaskAttemptsInfo(task_index) for task_index in range(self._job.total_tasks())]
        for attempt_id, status_update in tasks_status.items():
            attempts_status[attempt_id.task_index].add(attempt_id, status_update)

        # All task should have been attempted at least once, with QUEUED status or above
        count_zero_attempts = sum(1 for ts in attempts_status if ts.attempt_count == 0)
        if count_zero_attempts > 0:
            raise Exception("Some tasks have no attempt at all!")

        return attempts_status



"""
Calculate physical memory & cost for AWS Lambda-based workers.

Important note re. Lambda billing: although this is not explicitly stated and subject to change, you are not charged for
the duration in which a cold-started Lambda loads up till the point when the actual handler is called -
meaning, all imports are "free"! this means that cold-started Lambdas mainly impact clock-time latency but typically
won't inflate cost to a similar degree. This is in line with how the task duration is measured w/o cold-start imports.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import math
import re
from frocket.common.metrics import MetricName, EnvironmentMetricsProvider, MetricData

logger = logging.getLogger(__name__)

# TODO backlog setup a recurring task to check for pricing changes, so this can be updated.
DEFAULT_PRICE_GB_SEC = 0.0000166667
REGION_PRICING = {
    "eu-south-1": 0.0000195172,  # Milan
    "me-south-1": 0.0000206667,  # Bahrain
    "ap-northeast-3": 0.00002153,  # Osaka
    "af-south-1": 0.0000221,  # Capetown
    "ap-east-1": 0.00002292  # Hong-kong
}
# Assume the actual run takes this amount of seconds more than what's been measured,
# e.g. time spent in decoding the task reqeust, and time still to spend on writing results (incl. these metrics...)
# to datastore.
LAMBDA_TIME_OVERHEAD = 0.008  # 8ms, a conservative value based on a few observations


class AwsLambdaMetricsProvider(EnvironmentMetricsProvider):
    def __init__(self, lambda_context):
        # See https://docs.aws.amazon.com/lambda/latest/dg/python-context.html
        assert lambda_context.__class__.__name__ == 'LambdaContext'
        self._lambda_context = lambda_context

        # What region are we in? figure out by the full ARN in the context
        # (ARN example: arn:aws:lambda:us-west-2:123456789012:function:my-function)
        arn_parts = lambda_context.invoked_function_arn.split(':')
        region = arn_parts[3]
        if re.match(r'\w+-\w+-\d+', region):
            self._region = region
        else:
            self._region = None
            logger.warning(f"Seems like an invalid region: '{region}' in ARN: {lambda_context.invoked_function_arn}, "
                           f"not calculating cost")

    def _memory_bytes(self):
        mem_bytes = int(self._lambda_context.memory_limit_in_mb) * (1024 ** 2)
        return MetricData(MetricName.MACHINE_MEMORY_BYTES, mem_bytes)

    def _cost_dollars(self, duration=None):
        if not duration or not self._region:
            return None

        # noinspection PyBroadException
        try:
            memory_gb = self._memory_bytes().value / (1024 ** 3)
            # Lambdas are currently billed in 1ms granularity, so rounding up
            rounded_duration = duration + LAMBDA_TIME_OVERHEAD
            rounded_duration = math.ceil(rounded_duration * 1000) / 1000

            gb_second_units = rounded_duration * memory_gb
            cost_per_unit = REGION_PRICING.get(self._region, DEFAULT_PRICE_GB_SEC)
            cost = gb_second_units * cost_per_unit
            message = \
                f"Cost: original duration: {duration: .4f} sec, rounded duration: {rounded_duration:.3f}, memory: " \
                f"{memory_gb}GB, GB/second units: {gb_second_units}, unit cost for region {self._region}: " \
                f"${cost_per_unit:.10f} => total run cost is ${cost:.10f}"
            logger.debug(message)
            return MetricData(MetricName.COST_DOLLARS, cost)
        except Exception:
            logger.exception("Failed calculating cost")
            return None



"""
HTTP API server for Funnel Rocket - manage datasets, get metadata, run queries.
For long-running operations, the server supports streaming updates with HTTP chunked encoding.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import Type, Callable, cast
import flask
from flask import Flask, request, jsonify, Response, stream_with_context, json, make_response
from prometheus_client.exposition import make_wsgi_app
from werkzeug.middleware.dispatcher import DispatcherMiddleware
from frocket.common.config import config
from frocket.common.dataset import DatasetInfo
from frocket.common.serializable import SerializableDataClass
from frocket.common.tasks.registration import RegisterArgs, BaseApiResult
from frocket.common.tasks.async_tracker import AsyncJobTracker
from frocket.common.validation.error import ValidationErrorKind
from frocket.invoker import invoker_api

config.init_logging()
logger = logging.getLogger(__name__)

ALLOW_ADMIN_ACTIONS = config.bool("apiserver.admin.actions")
RETURN_ERROR_DETAILS = config.bool("apiserver.error.details")
PRETTY_PRINT = config.bool("apiserver.prettyprint")
EXPORT_TO_PROMETHEUS = config.bool('metrics.export.prometheus')
DEFAULT_ERROR = 'Error'

app = Flask(__name__)
if PRETTY_PRINT:
    app.config['JSONIFY_PRETTYPRINT_REGULAR'] = True

if EXPORT_TO_PROMETHEUS:
    # Add prometheus WSGI middleware to route /metrics requests
    app.wsgi_app = DispatcherMiddleware(app=app.wsgi_app, mounts={'/metrics': make_wsgi_app()})


class ClientFacingError(Exception):
    """
    For purposefully throwing errors back to client.
    'public' errors are those known not to leak any sensitive information. Other errors are supressed from response
    if 'apiserver.error.details' config attribute is False.
    """
    def __init__(self, message: str, public_error: bool = False):
        self.message = message
        self.public_error = public_error

    @staticmethod
    def error_message_to_client(error, public_error=False) -> str:
        if public_error or RETURN_ERROR_DETAILS:
            return f"Error: {error}"
        else:
            return DEFAULT_ERROR

    def __str__(self):
        return self.error_message_to_client(self.message, public_error=self.public_error)


@app.errorhandler(ClientFacingError)
def user_facing_error_handler(error: ClientFacingError):
    logger.exception(error)
    return str(error), 500


@app.errorhandler(500)
def error500_handler(error):
    message = DEFAULT_ERROR
    # noinspection PyBroadException
    try:
        message = ClientFacingError.error_message_to_client(error, public_error=False)  # Assuming non-public error
    except Exception:
        logger.exception('Error while formatting full error response')
    logger.exception(error)
    return message, 500


def get_dataset_or_fail(dataset_name: str) -> DatasetInfo:
    dataset = invoker_api.get_dataset(dataset_name, throw_if_missing=False)
    if dataset:
        return dataset
    else:
        raise ClientFacingError(message=f"Dataset {dataset_name} not found", public_error=True)


def ensure_admin_enabled() -> None:
    if not ALLOW_ADMIN_ACTIONS:
        raise ClientFacingError(message="Not supported", public_error=True)


def dataclass_from_body(request, cls: Type[SerializableDataClass]) -> SerializableDataClass:
    try:
        return cls.from_json(request.data)
    except KeyError as ke:
        raise ClientFacingError(f"{ke} missing or invalid in body", public_error=True)
    except ValueError as ve:
        raise ClientFacingError(f"{ve}", public_error=True)


def make_api_response(result: BaseApiResult) -> flask.Response:
    status_code = 200 if result.success else 500
    return make_response(jsonify(result.to_api_response_dict()), status_code)


def bool_request_arg(name: str, default: bool = False) -> bool:
    v = request.args.get(name, None)
    if v is None:
        return default
    else:
        return v.strip().lower() == "true"


def run_streamable(sync_func: Callable[[], BaseApiResult] = None,
                   async_func: Callable[[], AsyncJobTracker] = None,
                   should_stream: bool = False) -> flask.Response:
    """
    Run functions that are either synchronous (normal HTTP response, full response returned when handling ends) or
    async. (status updates are returned, then the final response, via HTTP chunked encoding). For async functions,
    the argument is an already initialized AsyncJobTracker that can be polled to completion.
    """
    assert (should_stream and async_func) or \
           (not should_stream and sync_func)

    try:
        @stream_with_context
        def generate_stream(tracker):
            # TODO backlog use the newer AsyncJobStatus.generator()
            while True:
                update_available = tracker.wait()
                if not tracker.wait_time_remaining:
                    api_result = BaseApiResult(success=False, error_message="Request timed out")
                    break

                status = tracker.status
                if status.result:
                    api_result = status.result
                    break

                if update_available:
                    if status.task_counters:
                        simple_status = {
                            'stage': status.stage.value,
                            'message': status.message,
                            'tasks': {k.name: v for k, v in status.task_counters.items()},
                        }
                        yield json.dumps(simple_status) + '\n'  # Chunks are expected to be separated by blank lines
                        logger.debug(f"Streaming status: {simple_status}")

            yield json.dumps(api_result.to_api_response_dict()) + '\n'

        if not should_stream:
            job_result = sync_func()
            return make_api_response(job_result)
        else:
            # TODO backlog test handling exceptions in async run
            tracker = async_func()
            return Response(generate_stream(tracker), mimetype='application/json')

    except Exception as e:
        raise ClientFacingError(str(e), public_error=False)


@app.route('/datasets/register', methods=['POST'])
def register_dataset():
    ensure_admin_enabled()
    register_args = cast(RegisterArgs, dataclass_from_body(request, RegisterArgs))
    should_stream = bool_request_arg('stream')
    if should_stream:
        return run_streamable(should_stream=True, async_func=lambda: invoker_api.register_dataset_async(register_args))
    else:
        return run_streamable(should_stream=False, sync_func=lambda: invoker_api.register_dataset(register_args))


@app.route('/datasets/<dataset_name>/unregister', methods=['POST'])
def unregister_dataset(dataset_name: str):
    ensure_admin_enabled()
    should_force = bool_request_arg('force')
    result = invoker_api.unregister_dataset(dataset_name, should_force)
    return make_api_response(result)


@app.route('/datasets')
def list_datasets():
    datasets = [ds.to_api_response_dict() for ds in invoker_api.list_datasets()]
    return jsonify(datasets)


@app.route('/datasets/<dataset_name>/schema')
def get_dataset_schema(dataset_name: str):
    dataset = get_dataset_or_fail(dataset_name)
    want_full_schema = bool_request_arg('full')
    schema = invoker_api.get_dataset_schema(dataset, full=want_full_schema)
    return jsonify(schema.to_api_response_dict())


@app.route('/datasets/<dataset_name>/parts')
def get_dataset_parts(dataset_name: str):
    ensure_admin_enabled()
    dataset = get_dataset_or_fail(dataset_name)
    parts = invoker_api.get_dataset_parts(dataset)
    return jsonify(parts.to_api_response_dict())


def do_run_query(dataset_name: str, query: dict, should_stream: bool = False) -> Response:
    dataset = get_dataset_or_fail(dataset_name)
    validation_result = invoker_api.expand_and_validate_query(dataset, query)
    if not validation_result.success:
        is_public_error = validation_result.error_kind != ValidationErrorKind.UNEXPECTED
        raise ClientFacingError(message=f"Query validation failed: {validation_result.error_message}",
                                public_error=is_public_error)

    query = validation_result.expanded_query
    if should_stream:
        return run_streamable(should_stream=True,
                              async_func=lambda:  invoker_api.run_query_async(dataset, query,
                                                                              validation_result=validation_result))
    else:
        return run_streamable(should_stream=False,
                              sync_func=lambda: invoker_api.run_query(dataset, query,
                                                                      validation_result=validation_result))


@app.route('/datasets/<dataset_name>/query', methods=['POST'])
def run_query(dataset_name: str):
    query = request.json
    should_stream = bool_request_arg('stream')
    return do_run_query(dataset_name, query, should_stream)


@app.route('/datasets/<dataset_name>/empty-query')
def run_empty_query(dataset_name: str):
    """Returns basic stats (group and row count, etc.) over the dataset. A GET request since no query is passed."""
    query = {}
    should_stream = bool_request_arg('stream')
    return do_run_query(dataset_name, query, should_stream)



"""
Base types for metrics support.

This was designed with Prometheus conventions & requirements (which aligns pretty well with OpenMetrics, surprise
surprise), however these classes are generally useful with other systems. This is specifically true, IMHO,
since good Prometheus compatibility means stricter definitions, e.g always use the same label set for a
given metric (rather than free for all whenever you report it), specify the *unit* of the metric as suffix, limit the
cardinality of label values, etc.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import time
from abc import ABCMeta, abstractmethod
from enum import auto
from typing import Union, Dict, Optional, List, NamedTuple, Type, Set
from dataclasses import dataclass
from inflection import underscore
from frocket.common.serializable import AutoNamedEnum, EnumSerializableByName
from frocket.common.helpers.utils import memoize

logger = logging.getLogger(__name__)


class MetricLabelEnum(AutoNamedEnum):
    """
    Base class for easily setting labels to metrics - the label name being inferred from the enum class name,
    and the value from the member name. Label names and values are transformed to snake-case to match the understated
    Prometheus style.
    """
    # noinspection PyUnusedLocal
    def __init__(self, *args):
        self.__class__._set_label_name()
        self.label_value = self.name.lower()

    @classmethod
    def _set_label_name(cls):
        """Label name is based on the class name, so needs to be done once."""
        if hasattr(cls, 'label_name'):
            return
        key = cls.__name__
        # Remove suffix if exists. TODO backlog when moving to Python 3.9 use removesuffix()
        key = key[:-len('Label')] if key.endswith('Label') else key
        cls.label_name = underscore(key)


class ComponentLabel(MetricLabelEnum):
    """Label for specifying which system component is reporting a metric."""
    INVOKER = auto()
    WORKER = auto()
    APISERVER = auto()


class WorkerStartupLabel(MetricLabelEnum):
    """Serverless worker types have a cold-start/warm-start distinction, which is meaning when monitoring timing."""
    COLD = auto()
    WARM = auto()


class LoadFromLabel(MetricLabelEnum):
    """Whether part files are loaded from origin, or some of the possible cache layers."""
    SOURCE = auto()
    DISK_CACHE = auto()
    MEMORY_CACHE = auto()


class PartSelectMethodLabel(MetricLabelEnum):
    """For jobs supporting part selection by the worker side, we want to monitor whether the worker has successfully
    selected a preferred (locally cached) part, or whether it fell back to choosing a 'random' available part due to
    either having no relevant cached parts, or because they were already taken by another worker."""
    SET_BY_INVOKER = auto()
    SPECIFIC_CANDIDATE = auto()
    RANDOM_NO_CANDIDATES = auto()
    RANDOM_CANDIDATES_TAKEN = auto()


class JobTypeLabel(MetricLabelEnum):
    REGISTER = auto()
    QUERY = auto()


class MeasuredUnit(EnumSerializableByName):
    """Each metric has a unit name as its suffix, which impacts the actual metric type to be created, and make life
    easier for human operators."""
    # noinspection PyUnusedLocal
    def __init__(self, *args):
        super().__init__()
        self.suffix = f"_{self.name}"

    # COUNT is special in that we're measuring occurences of something (a natural number, using a histogram is N/A)
    COUNT = auto()
    SECONDS = auto()
    ROWS = auto()
    GROUPS = auto()
    BYTES = auto()
    DOLLARS = auto()
    # OTHER is assumed to be something that's not a 'count of things'
    OTHER = auto()

    @staticmethod
    def unitof(metric_name: str):
        for e in MeasuredUnit:
            if metric_name.endswith(e.suffix):
                return e
        return MeasuredUnit.OTHER


MetricLabel = Union[MetricLabelEnum, str]

SUCCESS_LABEL = 'success'
DATASET_LABEL = 'dataset'


class MetricName(EnumSerializableByName):
    """All metrics must be defined as members and be associated with a single reporting component
    (to prevent confusion). Any additional metric-specific labels should also be defined at creation."""
    def __init__(self,
                 component: ComponentLabel,
                 unit: MeasuredUnit = None,
                 additional_labels: List[MetricLabel] = None):
        super().__init__()
        self.component = component
        self.unit = unit or MeasuredUnit.unitof(self.name)
        self.additional_labels = additional_labels

    INVOKE_TO_RUN_SECONDS = ComponentLabel.WORKER
    TASK_TOTAL_RUN_SECONDS = ComponentLabel.WORKER
    TASK_TOTAL_LOAD_SECONDS = ComponentLabel.WORKER
    TASK_PREFLIGHT_SLEEP_SECONDS = ComponentLabel.WORKER
    TASK_DOWNLOAD_SECONDS = ComponentLabel.WORKER
    TASK_LOAD_FILE_SECONDS = ComponentLabel.WORKER
    TASK_RUN_QUERY_SECONDS = ComponentLabel.WORKER
    SCANNED_ROWS = ComponentLabel.WORKER
    SCANNED_GROUPS = ComponentLabel.WORKER
    MACHINE_MEMORY_BYTES = ComponentLabel.WORKER
    COST_DOLLARS = ComponentLabel.WORKER
    INVOKER_TOTAL_SECONDS = ComponentLabel.INVOKER
    ASYNC_ENQUEUE_SECONDS = ComponentLabel.INVOKER
    ASYNC_POLL_SECONDS = ComponentLabel.INVOKER


# Main metric for work duration (scope is a job for the invoker, a single task in a job for the worker)
COMPONENT_DURATION_METRIC: Dict[ComponentLabel, MetricName] = {
    ComponentLabel.WORKER: MetricName.TASK_TOTAL_RUN_SECONDS,
    ComponentLabel.INVOKER: MetricName.INVOKER_TOTAL_SECONDS
}

# There's a standard common set of labels for *all* metrics, then another set depending on the component.
# TODO backlog rethink this set: on one hand, having a common set allows filtering by label values across metrics
#  (thus preventing mismatches when looking at graphs). On the other, this inflates the 'cardinality' of metrics by
#  a lot (each extra labels with, say, 3 possible values, means up to x3 the no. of unique label sets that Prometheus
#  would need to track per metrics. Perhaps this should be configurable...
ALL_COMPONENT_COMMON_LABELS: List[MetricLabel] = [ComponentLabel, SUCCESS_LABEL, DATASET_LABEL, JobTypeLabel]

COMPONENT_COMMON_LABELS: Dict[ComponentLabel, List[MetricLabel]] = {
    ComponentLabel.WORKER: [*ALL_COMPONENT_COMMON_LABELS, WorkerStartupLabel, LoadFromLabel, PartSelectMethodLabel],
    ComponentLabel.INVOKER: [*ALL_COMPONENT_COMMON_LABELS]
}

# Label values can be accepted as numbers, but are then stored as text (which is how metric systems get them)
LabelsDict = Dict[str, str]


def label_to_str(label: Union[MetricLabel, Type[MetricLabelEnum]]) -> str:
    """Given either a string, a MetricLabel enum *member or class* - return the label name"""
    if type(label) is str:
        return label
    elif isinstance(label, MetricLabelEnum):
        return label.label_name
    elif isinstance(label, type) and issubclass(label, MetricLabelEnum):
        return label.label_name
    else:
        raise Exception(f"Invalid label: {label} ({type(label)})")


@memoize
def supported_label_names(metric: MetricName) -> List[str]:
    all_labels = COMPONENT_COMMON_LABELS[metric.component] + (metric.additional_labels or [])
    return [label_to_str(label) for label in all_labels]


@memoize
def empty_label_names(metric: MetricName) -> Dict[str, Union[str, None]]:
    return {label: None for label in supported_label_names(metric)}


ALL_LABEL_NAMES: Set[str] = set()
for m in MetricName:
    ALL_LABEL_NAMES.update(supported_label_names(m))


@dataclass(frozen=True)
class MetricData:
    """A instance of a reported metric - name, value and optional metric-specific labels."""
    name: MetricName
    value: float
    labels: Optional[LabelsDict] = None

    def with_added_labels(self, added_labels: LabelsDict):
        """Add "inherited" labels from the component level/all components level."""
        if self.labels:
            for k in added_labels.keys():
                if self.labels.get(k, None) is not None:
                    raise Exception(f"Trying to override existing initialized labels. "
                                    f"Metric: {self}, added labels: {added_labels}")
            merged_labels = {**self.labels, **added_labels}
        else:
            merged_labels = added_labels
        return MetricData(name=self.name, value=self.value, labels=merged_labels)


class SourceAndMetricTuple(NamedTuple):
    """For passing along a metric with the specific source which reported it (e.g. specific task names)."""
    source: str
    metric: MetricData


class EnvironmentMetricsProvider(metaclass=ABCMeta):
    """A base class for providing metrics relevant to the runtime environment (AWS Lambda, generic processes,
    and future runtimes), used by the worker. Not all providers may support all possible metrics."""
    def finalize_metrics(self, run_duration: float = None) -> List[MetricData]:
        metrics = [self._memory_bytes(), self._cost_dollars(run_duration)]
        return [m for m in metrics if m]

    @abstractmethod
    def _memory_bytes(self) -> Optional[MetricData]:
        """Runtime environment-dependent implementation for total physical memory (whether used or free"""
        pass

    @abstractmethod
    def _cost_dollars(self, duration: float = None) -> Optional[MetricData]:
        """environment-specific approximate cost calculation - if possible to get."""
        pass


class MetricsBag:
    """A MetricsBag object accompanies a job (at the invoker level) or a task (at the worker level) along its lifecycle,
    being propagated to all relevant modules through the process. Metrics & common labels are reported through it."""
    def __init__(self, component: ComponentLabel, env_metrics_provider: EnvironmentMetricsProvider = None):
        self._component: ComponentLabel = component
        self._metrics: Dict[MetricName, MetricData] = {}
        self._env_metrics_provider = env_metrics_provider
        self._finalized = False
        self._labels: LabelsDict = {}  # Labels (value->name) common to all metrics in this bag, rather than one metric
        self.set_label_enum(component)

    def set_label(self, name: str, value: Union[str, int, float, bool]):
        self._do_set_label(name, value)

    def set_label_enum(self, e: MetricLabelEnum):
        """Easier, less error-prone method for setting labels - just pass the appropriate enum member."""
        self._do_set_label(e, e.label_value)

    def _do_set_label(self, label: MetricLabel, value):
        self.fail_if_finalized()
        string_name = label_to_str(label)
        lookup = label if type(label) is str else type(label)
        if lookup not in COMPONENT_COMMON_LABELS[self._component]:
            raise Exception(f"Label {label} ({type(label)}) not supported for {self._component}")

        if string_name in self._labels:
            raise Exception(f"Label '{string_name}' already set for {self._component}")
        self._labels[string_name] = str(value)

    def set_metric(self, metric: MetricName, value: float,
                   additional_labels: Dict[MetricLabel, str] = None):
        self.fail_if_finalized()
        if metric.component != self._component:
            raise Exception(f"Metric {metric} not applicable for {self._component}")
        elif metric in self._metrics:
            raise Exception(f"Metric {metric} already set")
        elif additional_labels:
            for k in additional_labels.keys():
                if k not in metric.additional_labels:
                    raise Exception(f"Label {k} not applicable for {metric}")

        data = MetricData(metric, value, additional_labels)
        self._metrics[metric] = data

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Set {data}")

    def measure(self, name: MetricName):
        """Start a 'context manager' for measuring a block of code to the given metric. When the code block ends,
        its duration will be automatically added to this bag."""
        return MeasuredBlock(self, name)

    @property
    def success(self) -> Optional[bool]:
        return self._labels.get(SUCCESS_LABEL, None)

    def finalize(self, success: bool, added_labels: LabelsDict = None) -> List[MetricData]:
        """Prepare metrics for shipping (from worker back to invoker, through the data store). No further metrics
        can be collected on this bag, for consistency."""
        self.fail_if_finalized()
        self.set_label(SUCCESS_LABEL, success)
        self._add_env_metrics()
        self._finalized = True

        # The returned metrics get all inherited labels from this bag
        common_labels = {**(added_labels or {}),
                         **self._labels}
        result = [m.with_added_labels(common_labels)
                  for m in self._metrics.values()]
        return result

    def _add_env_metrics(self):
        if not self._env_metrics_provider:
            return

        duration_metric = self._metrics.get(COMPONENT_DURATION_METRIC[self._component], None)
        run_duration = duration_metric.value if duration_metric else None
        env_metrics = self._env_metrics_provider.finalize_metrics(run_duration)
        self._metrics.update({m.name: m for m in env_metrics})

    def fail_if_finalized(self):
        if self._finalized:
            raise Exception("Already finalized")

    def __getitem__(self, metric: MetricName) -> Optional[float]:
        if metric.name in self._metrics:
            return self._metrics[metric].value
        else:
            return None

    def label_value(self, label: Union[MetricLabel, Type[MetricLabelEnum]]):
        return self._labels.get(label_to_str(label), None)


class MeasuredBlock:
    """
    See MetricsBag.measure() method above: measure a code block and report its duration back to the parent bag,
    as the given metric's value.
    TODO backlog consider capturing any exceptions in the code block before re-raising them,
     so a success=False label can be applied to this metric.
    """
    def __init__(self, parent: MetricsBag, name: MetricName):
        self._parent = parent
        self._name = name
        self._start_time = time.time()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        total_time = time.time() - self._start_time
        self._parent.set_metric(self._name, total_time)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import json
import logging
import sys
from typing import List, Union
import difflib
import jsonschema
from frocket.common.helpers.utils import terminal_green, terminal_red
from frocket.common.validation.result import QueryValidationResult
from frocket.common.validation.error import ValidationErrorKind, QueryValidationError
from frocket.datastore.registered_datastores import get_datastore
from frocket.common.dataset import DatasetColumnType, DatasetInfo, DatasetShortSchema
from frocket.common.validation.consts import QUERY_SCHEMA, AGGREGATIONS_PATHS, SINGLE_FILTER_PATHS, \
    FILTER_ARRAY_PATHS, VALID_IDENTIFIER_PATTERN, UNIQUE_IDENTIFIER_SCOPES, OPERATORS_BY_COLTYPE, \
    VALUE_TYPES_BY_COLTYPE, NUMERIC_COLTYPES, RELATION_OPS, DEFAULT_RELATION_OP, CONDITION_COLUMN_PREFIX, \
    map_condition_names
from frocket.common.validation.relation_parser import RelationParser
from frocket.common.validation.visitor_functions import _to_verbose_filter, _to_verbose_target, _add_default_target, \
    _validate_aggregation, _expand_aggregations, _validate_or_set_include_zero
from frocket.common.validation.path_visitor import PathVisitor, PathVisitorCallback
from frocket.engine.relation_to_pandas import relation_to_pandas_query

logger = logging.getLogger(__name__)


# TODO doc asserts
class QueryValidator:
    """
    Validate a query, and *expand* as needed:

    The query schema as documented externally allows for various shorthand notations and for some "sensible" defaults
    to be ommitted. However, the query engine itself expects the full notation. To hide this complexity from the engine,
    all transformations are done here, and the 'expanded' query is returned in the result -
    that's the version that should be passed to query tasks rather than the user-supplied one.

    'assert' is used in various places in code where some conditions are expected to be already validated, but if not -
    we want the code to crash.
    """

    def __init__(self, source_query: dict, dataset: DatasetInfo = None, short_schema: DatasetShortSchema = None):
        """Unless calling expand_and_validate() in schema_only=True more, dataset and its schema must be passed."""
        self._source_query = source_query
        self._expanded_query = None
        self._dataset = dataset
        self._short_schema = short_schema
        self._used_columns = None
        self._condition_mapping = None
        self._used_conditions = None
        self._relation_elements = None

    def expand_and_validate(self, schema_only: bool = False) -> QueryValidationResult:
        """
        Validates and applies any transformations.
        In schema_only=True mode, only 'static' validation is done, meaning the existence of columns and their types
        is not checked against the actual dataset the user wishes to query.
        """
        try:
            if not schema_only and not (self._dataset and self._short_schema):
                raise QueryValidationError(
                    message="Provide a dataset and its short schema, or use schema_only=True",
                    kind=ValidationErrorKind.INVALID_ARGUMENTS)

            # First, validate using the JSON schema.
            try:
                jsonschema.validate(instance=self._source_query, schema=QUERY_SCHEMA)
            except jsonschema.exceptions.ValidationError as ve:
                raise QueryValidationError(message=ve.message, kind=ValidationErrorKind.SCHEMA)

            # Second, perform all validations & expansions which do not need the dataset details.
            # Clone before making changes (using serialization, which may also catch unexpected types in the dict)
            self._expanded_query = json.loads(json.dumps(self._source_query))
            self._validate_and_expand_schema()

            # Third, perform validations given the specific dataset and its schema
            if not schema_only:
                self._validate_columns()
                self._validate_timeframe(dataset_mintime=self._short_schema.min_timestamp,
                                         dataset_maxtime=self._short_schema.max_timestamp)

            return QueryValidationResult(success=True,
                                         source_query=self._source_query,
                                         expanded_query=self._expanded_query,
                                         used_columns=self._used_columns,
                                         used_conditions=self._used_conditions,
                                         named_conditions=self._condition_mapping.names,
                                         relation_elements=self._relation_elements,
                                         warnings=None)  # TODO backlog: support warnings/hints to the user
        except Exception as e:
            if not isinstance(e, QueryValidationError):
                logger.exception("Unexpected error")
            return QueryValidationResult.from_exception(e, self._source_query)

    def _validate_and_expand_schema(self):
        # Convert shorthand filter notation (a list) to verbose notation (dict) in single-filter conditions
        # and in sequence steps (both under sequence conditions or funnel).
        # (in filter arrays, filters are defined in schema to always be in verbose notation)
        self._modify(paths=SINGLE_FILTER_PATHS, list_to_items=False, func=_to_verbose_filter)

        # Convert shorthand target notation (a list) to verbose (dict)
        self._modify(paths='query.conditions.target', list_to_items=False, func=_to_verbose_target)

        # Put default target in conditions where it's omitted
        self._visit(paths='query.conditions', func=_add_default_target)

        # Expand 'default' column aggregation (with no declared type) to actual list of aggregations
        self._modify(paths=AGGREGATIONS_PATHS, list_to_items=False, func=_expand_aggregations)

        # Check value of 'includeZero' optional flag, and its applicability. Explicitly set where needed.
        self._visit(paths='query.conditions', func=_validate_or_set_include_zero)

        # Validate that only aggregation requiring otherColumn have them
        self._visit(paths=AGGREGATIONS_PATHS, func=_validate_aggregation)

        # Ensure uniquene names for condition column aggregations
        for scope in UNIQUE_IDENTIFIER_SCOPES:
            self._validate_unique_identifiers(scope)

        # Get total condition count, map named condition to indexes
        self._validate_condition_names()
        # Set a default relation if omitted, expand a shorthand relation,
        # and then fully parse and transform the relation
        self._expand_and_parse_relation()

        # Validate optional timeframe, first without dataset boundaries
        self._validate_timeframe()

        # Self-test that our own modifications did not trash the schema somehow
        jsonschema.validate(instance=self._expanded_query, schema=QUERY_SCHEMA)

    def _visit(self, paths: Union[str, List[str]],
               func: PathVisitorCallback,
               modifiable: bool = False,
               list_to_items: bool = True):
        if type(paths) is str:
            paths = [paths]
        for path in paths:
            p = PathVisitor(root=self._expanded_query, path=path, modifiable=modifiable, list_to_items=list_to_items)
            p.visit(func)

    def _modify(self, paths: Union[str, List[str]], list_to_items: bool, func: PathVisitorCallback) -> None:
        self._visit(paths, modifiable=True, list_to_items=list_to_items, func=func)

    def _collect(self, paths: Union[str, List[str]], list_to_items: bool = True) -> list:
        results = []
        self._visit(paths, modifiable=False, list_to_items=list_to_items, func=lambda v: results.append(v))
        return results

    def _validate_unique_identifiers(self, scope: str) -> None:
        def add_identifier(v):
            if v:
                if v in seen_values:
                    raise QueryValidationError(message=f"Identifier name '{v}' is not unique in scope '{scope}'",
                                               kind=ValidationErrorKind.SCHEMA)
                elif not VALID_IDENTIFIER_PATTERN.match(v):
                    raise QueryValidationError(message=f"Invalid identifier name '{v}'",
                                               kind=ValidationErrorKind.SCHEMA)
                seen_values.add(v)

        seen_values = set()
        self._visit(paths=scope, modifiable=False, list_to_items=True, func=add_identifier)

    def _validate_columns(self):
        # As a baseline, these columns are always needed for the query to work
        used_columns = [self._dataset.group_id_column, self._dataset.timestamp_column]

        # Collect all filter objects and validate:
        # (a) column existence, (b) suitability of operator and type to the column type
        all_filters = self._collect(paths=SINGLE_FILTER_PATHS, list_to_items=False) + \
            self._collect(paths=FILTER_ARRAY_PATHS, list_to_items=True)

        for f in all_filters:
            self._validate_column(name=f['column'], op=f['op'], value=f['value'])
            used_columns.append(f['column'])

        # Collect column names from aggregations
        for name in self._collect(paths=[f"{path}.column" for path in AGGREGATIONS_PATHS]):
            self._validate_column(name)
            used_columns.append(name)

        # Collect column name referenced in (a) targets of type 'sum' and (b) aggregations of type 'sumPerValue',
        # these must be numeric columns.
        sum_by_column_paths = ["query.conditions.target.column"] + \
                              [f"{path}.otherColumn" for path in AGGREGATIONS_PATHS]
        sum_by_columns = self._collect(paths=sum_by_column_paths)
        for name in sum_by_columns:
            self._validate_column(name=name, expected_types=NUMERIC_COLTYPES)
            used_columns.append(name)

        self._used_columns = list(set(used_columns))

    def _validate_column(self,
                         name: str,
                         expected_types: List[DatasetColumnType] = None,
                         op: str = None,
                         value: Union[int, float, str, bool] = None):
        coltype = self._short_schema.columns.get(name, None)
        if not coltype:
            raise QueryValidationError(f"Column '{name}' not in dataset", kind=ValidationErrorKind.DATASET_MISMATCH)

        if expected_types and coltype not in expected_types:
            raise QueryValidationError(f"Column '{name}' is of type {coltype.value}, but should be one of "
                                       f"{[t.value for t in expected_types]}", kind=ValidationErrorKind.TYPE_MISMATCH)

        if op:
            if op not in OPERATORS_BY_COLTYPE[coltype]:
                raise QueryValidationError(f"Operator '{op}' is not applicable for column '{name}' "
                                           f"of type {coltype.value}", kind=ValidationErrorKind.TYPE_MISMATCH)
        if value is not None:
            assert op  # Why was op not passed?
            if type(value) not in VALUE_TYPES_BY_COLTYPE[coltype]:
                value_as_str = (f'"{value}"' if type(value) is str else str(value)) + \
                               f" (type {type(value).__name__.upper()})"
                raise QueryValidationError(f"Value {value_as_str} "
                                           f"is not applicable for column '{name}' of type {coltype.value}",
                                           kind=ValidationErrorKind.TYPE_MISMATCH)

    def _validate_condition_names(self):
        self._condition_mapping = map_condition_names(self._expanded_query)
        for name in self._condition_mapping.names.keys():
            if name in RELATION_OPS:
                raise QueryValidationError(f"'{name}' is not a valid condition name",
                                           kind=ValidationErrorKind.SCHEMA)

    def _expand_and_parse_relation(self):
        relation = None
        simple_op = None
        found_relations = self._collect('query.relation')
        assert len(found_relations) in [0, 1]

        if not found_relations:
            if not self._collect('query.conditions'):
                return
            else:  # There are conditions under 'query', but no relation was defined - set default one
                simple_op = DEFAULT_RELATION_OP
                self._expanded_query['query']['relation'] = simple_op
        else:
            relation = found_relations[0].strip().lower()
            if relation in RELATION_OPS:
                simple_op = relation

        if simple_op:
            all_condition_ids = [f"${i}" for i in range(self._condition_mapping.count)]
            relation = f" {simple_op} ".join(all_condition_ids)
            self._expanded_query['query']['relation'] = relation

        if len(relation) == 0:
            raise QueryValidationError("Relation cannot be an empty string", kind=ValidationErrorKind.RELATION)

        parser = RelationParser(self._expanded_query)
        self._relation_elements = parser.parse()
        # TODO return a warning on unused conditions (when doing warnings mechanism)
        self._used_conditions = parser.used_conditions

    def _validate_timeframe(self, dataset_mintime: float = None, dataset_maxtime: float = None):
        """
        If a timeframe object is included in the query (either from, to, or both), validate that to > from,
        but also that the timestamps seem to be consistent with the dataset's timestamps.
        Currently, Funnel Rocket itself is not opinionated re. the timestamp resolution (seconds? milliseconds? second
        fractions as decimal places after the dot?), and only validates that the no. of digits seems ok.
        TODO backlog (optional?) strict timestamp specification
        TODO backlog when warnings are supported, warn client when query timeframe is fully outside dataset timeframe
        """
        def validate_scale(ts_in_query, ts_in_dataset, dataset_ts_name: str):
            if len(str(int(ts_in_query))) != len(str(int(ts_in_dataset))):
                message = f"Given timestamp {ts_in_query} doesn't appear to be in same scale as " \
                          f"{dataset_ts_name} ({ts_in_dataset}). Query timeframe is: {timeframe}"
                raise QueryValidationError(message, kind=ValidationErrorKind.TYPE_MISMATCH)

        timeframe = self._expanded_query.get('timeframe', None)
        if not timeframe:
            return

        fromtime = timeframe.get('from', None)
        totime = timeframe.get('to', None)

        if fromtime is not None and dataset_mintime is not None:
            validate_scale(fromtime, dataset_mintime, 'minimum timestamp in dataset')
        if totime is not None and dataset_maxtime is not None:
            validate_scale(totime, dataset_maxtime, 'maximum timestamp in dataset')

        if fromtime is not None and totime is not None:
            if totime <= fromtime:
                raise QueryValidationError(f"Value of 'to' (exclusive) should be larger than 'from' (inclusive) "
                                           f"in query timeframe: {timeframe}",
                                           kind=ValidationErrorKind.SCHEMA)

    def _fail_if_not_ran(self):
        if not self._expanded_query:
            raise Exception("Run validation first")

    def diff_from_source(self, colorize: bool = False) -> List[str]:
        """Handy utility for manual tests: show a nice diff between given and expanded queries, just like Git."""
        self._fail_if_not_ran()
        diff = difflib.unified_diff(
            a=json.dumps(self._source_query, indent=2).splitlines(),
            b=json.dumps(self._expanded_query, indent=2).splitlines(),
            fromfile="Source query",
            tofile="Expanded query")

        lines = []
        for curr_line in diff:
            line = curr_line
            if colorize:
                if line.startswith('+ '):
                    line = terminal_green(line)
                elif line.startswith('- '):
                    line = terminal_red(line)
            lines.append(line)
        return lines


# For running validations manually, including detailed diff's (source vs. expanded query)
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Manually run QueryValidator')
    parser.add_argument('dataset', type=str, help='Dataset name')
    parser.add_argument('query_file', type=str, help='Path to file with the query to validate')
    parser.add_argument('--diff', action='store_true', help='Show diff between original and expanded query')

    args = parser.parse_args()
    source_query = json.load(open(args.query_file, 'r'))

    store = get_datastore()
    dataset = store.dataset_info(args.dataset)
    short_schema = store.short_schema(dataset)

    validator = QueryValidator(source_query, dataset, short_schema)
    result = validator.expand_and_validate()
    if not result.success:
        sys.exit(f"{terminal_red('Validation failed!')}\n{result.error_message}")

    print(terminal_green("Validation successful!"))
    if args.diff:
        diff = validator.diff_from_source(colorize=True)
        for line in diff:
            print(line)

    print("===========================================================================================\n"
          f"Used columns: {result.used_columns}")
    print(f"Named conditions: {result.named_conditions}")
    print(f"Used conditions: {result.used_conditions}")
    print(f"Original relation: {PathVisitor(source_query, 'query.relation').list()[0]}")
    pandas_relation = \
        relation_to_pandas_query(elements=result.relation_elements, column_prefix=CONDITION_COLUMN_PREFIX)
    print(f"...........parsed: {pandas_relation}")



"""
Base class for running a task in a worker - to be subclassed for concerete task runners.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import time
from abc import abstractmethod
from typing import Optional
from frocket.common.config import config
from frocket.common.metrics import MetricName, MetricsBag
from frocket.common.tasks.base import TaskStatus, BaseTaskRequest, BaseTaskResult, TaskAttemptId
from frocket.datastore.datastore import Datastore
from frocket.datastore.blobstore import Blobstore
from frocket.datastore.registered_datastores import get_datastore, get_blobstore
from frocket.worker.runners.part_loader import PartLoader, shared_part_loader

logger = logging.getLogger(__name__)
REQUEST_MAX_AGE = int(config.get("worker.reject.age"))
DEFAULT_PREFLIGHT_DURATION_MS = config.int("part.selection.preflight.ms")


class TaskRunnerContext:
    """simple dependency provider... (for easier testing)."""
    def __init__(self,
                 metrics: MetricsBag,
                 private_part_loader: PartLoader = None,
                 preflight_duration_ms: int = None):
        self._metrics = metrics
        # By default, files are loaded and cached by a re-usable loaded.
        # Having a 'private' one allows testing in isolation
        self._part_loader = private_part_loader or shared_part_loader()
        if preflight_duration_ms is None:
            preflight_duration_ms = DEFAULT_PREFLIGHT_DURATION_MS
        self._preflight_duration_seconds = preflight_duration_ms / 1000

    @property
    def metrics(self) -> MetricsBag:
        return self._metrics

    # The underlying get_datastore and get_blobstore are memoized - initialized on demand
    @property
    def datastore(self) -> Datastore:
        return get_datastore()

    @property
    def blobstore(self) -> Blobstore:
        return get_blobstore()

    @property
    def part_loader(self) -> PartLoader:
        return self._part_loader

    @property
    def preflight_duration_seconds(self) -> float:
        return self._preflight_duration_seconds


class BaseTaskRunner:
    # Returns (should_run, reject_reason)
    @classmethod
    def should_run(cls, req: BaseTaskRequest) -> (bool, str):
        if cls.time_since_invocation(req) > REQUEST_MAX_AGE:
            return False, f"request is more than {REQUEST_MAX_AGE} seconds old"
        else:
            return True, None

    @staticmethod
    def time_since_invocation(req: BaseTaskRequest):
        return time.time() - req.invoke_time

    def __init__(self, req: BaseTaskRequest,
                 ctx: TaskRunnerContext):
        self._req = req
        self._ctx = ctx
        # TODO backlog initialize the attempt_id on init, if available (n/a here in self-select part mode)
        self._task_attempt_id: Optional[TaskAttemptId] = None

    def run(self) -> BaseTaskResult:
        error_message, engine_result = None, None
        with self._ctx.metrics.measure(MetricName.TASK_TOTAL_RUN_SECONDS):
            try:
                self._ctx.metrics.set_metric(MetricName.INVOKE_TO_RUN_SECONDS,
                                             self.time_since_invocation(self._req))

                self._do_run()  # Call concrete class to do the actual work
                final_status = TaskStatus.ENDED_SUCCESS
            except Exception as e:
                final_status = TaskStatus.ENDED_FAILED
                error_message = str(e)
                logger.exception('Task FAILED!')

        # Post-run: extracting the task metrics, building the concrete result object
        final_metrics = self._ctx.metrics.finalize(success=(final_status == TaskStatus.ENDED_SUCCESS))
        # First, set the base attributes in a dict as kind of a 'skeleton' response - then pass it to the concrete
        # task runner to pass as **args to the concrete result class
        base_attributes = BaseTaskResult(
            task_index=self._task_attempt_id.task_index,
            status=final_status,
            error_message=error_message,
            metrics=final_metrics).shallowdict(include_none=True)
        result = self._build_result(base_attributes)  # Call concrete class

        # If the job failed to get a task attempt ID assigned to it (self-select failed),
        # or if the datastore is not available - task status and result cannot be written
        # TODO backlog consider having an optional secondary channel to report such failures
        #  (aside from centralized logging?)
        if self._task_attempt_id:
            self._ctx.datastore.write_task_result(self._req.request_id, self._task_attempt_id, result)
        else:
            logger.error("Can't report result: no part was selected for loading")

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(result)
        return result

    def _update_status(self, status: TaskStatus):
        self._ctx.datastore.update_task_status(self._req.request_id, self._task_attempt_id, status)

    @abstractmethod
    def _do_run(self):
        pass

    @abstractmethod
    def _build_result(self, base_attributes: dict):
        """This method is still called by run() above even if _do_run() has raised an exception - having a sane
        result object is important even if a failed one."""
        pass



"""
Simple abstraction of local & remote filesystems.

Currently supports either a local filesystem (for non-distributed usage, or potentially a fast network share)
and S3 (or S3-compatible object stores such as MinIO, which is used for running tests).
Additional protocols are welcome.

TODO backlog: support pagination for S3 listing (so more than 1,000 files per dataset)
TODO backlog: support auto-identification of numbering pattern in dataset files, so the full list of filenames
 would not have to reside in the datastore
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import re
import tempfile
import uuid
from abc import abstractmethod
from enum import Enum, auto
from fnmatch import fnmatch
from pathlib import Path
from typing import NamedTuple, Optional, List
import boto3
from frocket.common.config import config
from frocket.common.dataset import DatasetPartsInfo, PartNamingMethod

logger = logging.getLogger(__name__)


class StorageHandler:
    """Simple abstraction of a storage protocol."""
    class FileBaseInfo(NamedTuple):
        relpath: str
        size: int

    def __init__(self, path: str):
        assert self.valid(path)
        self._path = path

    @classmethod
    def valid(cls, path: str) -> bool:
        """For validation of a path prior to instantiating the handler - a nicety instead of exceptions later,
        to be overriden where appropriate."""
        return True

    @property
    @abstractmethod
    def remote(self) -> bool:
        """This affects the caching behavior used by workers (see part_loader.py)."""
        pass

    @abstractmethod
    def _list_files(self, pattern: str) -> List[FileBaseInfo]:
        """Override in subclasses"""
        pass

    def discover_files(self, pattern: str) -> DatasetPartsInfo:
        files = self._list_files(pattern)
        files.sort(key=lambda fi: fi.relpath)
        # TODO backlog implement PartNamingMethod.RUNNING_NUMBER for compact metadata in large datasets
        parts_info = DatasetPartsInfo(naming_method=PartNamingMethod.LIST,
                                      total_parts=len(files),
                                      total_size=sum([fi.size for fi in files]),
                                      filenames=[fi.relpath for fi in files],
                                      running_number_pattern=None)
        return parts_info

    @abstractmethod
    def _local_path(self, fullpath: str) -> str:
        """
        If the filesystem is remote, download and return a local copy.
        Files should be cleaned-up by the caller which controls the caching behavior.
        """
        pass

    def get_local_path(self, fullpath: str) -> str:
        if not fullpath.startswith(self._path):
            raise Exception(f"Given full path {fullpath} is not under handler's path {self._path}")

        return self._local_path(fullpath)


class FileStorageHanler(StorageHandler):
    """Super-simple local filesystem handler"""
    @property
    def remote(self):
        return False

    def _list_files(self, pattern):
        paths = Path(self._path).iterdir()
        files = [StorageHandler.FileBaseInfo(path.name, path.stat().st_size)
                 for path in paths
                 if fnmatch(path.name, pattern)]
        return files

    def _local_path(self, fullpath):
        if not Path(fullpath).is_file():
            raise Exception(f"Path is missing/not a file: {fullpath}")
        return fullpath


class S3StorageHanler(StorageHandler):
    """S3 filesystem handler, supports datasets directly under the bucket or within a sub-directory."""
    S3_PATH_REGEX = re.compile(r"^s3://([a-zA-Z0-9_\-.]+)/([a-zA-Z0-9_\-./]*)$")

    def __init__(self, path: str):
        super().__init__(path)
        path_parts = self.S3_PATH_REGEX.match(path)
        self._bucket = path_parts.group(1)
        self._path_in_bucket = path_parts.group(2)
        no_trailing_slash = self._path_in_bucket and self._path_in_bucket[-1:] != '/'
        self._path_in_bucket_normalized = self._path_in_bucket + ('/' if no_trailing_slash else '')

    @classmethod
    def valid(cls, path):
        return True if cls.S3_PATH_REGEX.match(path) else False

    @property
    def remote(self):
        return True

    def _list_files(self, pattern):
        path_in_bucket = self._path_in_bucket_normalized
        logger.info(f"Listing files in S3 with bucket {self._bucket} and prefix {path_in_bucket}...")
        # TODO backlog support pagination
        s3response = self._client().list_objects_v2(Bucket=self._bucket, Prefix=path_in_bucket)

        filename_start_idx = len(path_in_bucket)
        path_to_size = {obj['Key'][filename_start_idx:]: obj['Size'] for obj in s3response['Contents']}
        files = [StorageHandler.FileBaseInfo(path, size)
                 for path, size in path_to_size.items()
                 if fnmatch(path, pattern)]
        return files

    def _local_path(self, fullpath):
        localpath = str(Path(tempfile.gettempdir()) / str(uuid.uuid4()))
        logger.info(f"Downloading {fullpath} to {localpath}...")
        self._client().download_file(self._bucket, self._path_in_bucket, localpath)
        return localpath

    @classmethod
    def _client(cls):
        if not hasattr(cls, '_s3client'):
            cls._s3client = boto3.client('s3', **config.aws_client_settings(service='s3'))
        return cls._s3client


class StorageProtocol(Enum):
    FILE = auto()
    S3 = auto()

    @classmethod
    def get(cls, name: str):
        return cls.__members__.get(name.upper())

    @classmethod
    def names(cls) -> List[str]:
        return list(cls.__members__.keys())


PATH_WITH_PROTOCOL_RE = r'(\w+)://(.+)$'
PROTOCOL_TO_HANDLER = {
    StorageProtocol.FILE: FileStorageHanler,
    StorageProtocol.S3: S3StorageHanler
}


def storage_handler_for(path: str, throw_if_missing: bool = True) -> Optional[StorageHandler]:
    """
    Instantiate the appropriate handler for the given path.
    Paths without explicit protocol are considered local.
    """
    path_and_protocol = re.match(PATH_WITH_PROTOCOL_RE, path)
    if path_and_protocol:
        method_name = path_and_protocol.groups()[0]
        method = StorageProtocol.get(method_name)
        if not method:
            if throw_if_missing:
                raise Exception(f"Storage protocol '{method_name}' is not in supported list: {StorageProtocol.names()}")
            else:
                return None
        elif method == StorageProtocol.FILE:
            path = path_and_protocol.groups()[1]
    else:
        method = StorageProtocol.FILE

    handler_cls = PROTOCOL_TO_HANDLER[method]
    if not handler_cls.valid(path):
        raise Exception(f"Invalid path: {path} (protocol: {method.name})")
    return handler_cls(path)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import tempfile
from typing import List, Type
from frocket.common.metrics import MetricName, MetricData, MetricLabelEnum

SKIP_SLOW_TESTS = os.environ.get('SKIP_SLOW_TESTS', "False").lower() == 'true'
SKIP_LAMBDA_TESTS = os.environ.get('SKIP_LAMBDA_TESTS', "False").lower() == 'true'
# noinspection PyProtectedMember,PyUnresolvedReferences
TEMP_DIR = tempfile._get_default_tempdir()


# noinspection PyProtectedMember,PyUnresolvedReferences
def temp_filename(suffix='', with_dir: bool = True):
    fname = next(tempfile._get_candidate_names()) + suffix
    return f"{TEMP_DIR}/{fname}" if with_dir else fname


# A mixin to allow defining utility classes named "Test<X>" without pytest trying to collect test cases in them,
# which results in warnings (and without needing a pytest.ini entry). See https://stackoverflow.com/a/46199666
class DisablePyTestCollectionMixin(object):
    __test__ = False


def get_metric_value(metrics: List[MetricData], name: MetricName) -> float:
    assert metrics
    metric = next(filter(lambda metric: metric.name == name, metrics), None)
    assert metric is not None
    return metric.value


def assert_metric_value(metrics: List[MetricData], name: MetricName, value: float):
    assert get_metric_value(metrics, name) == value


def find_first_label_value(metrics: List[MetricData], label_type: Type[MetricLabelEnum]) -> str:
    assert metrics
    found_metric = next(filter(lambda metric: label_type.label_name in metric.labels, metrics), None)
    return found_metric.labels[label_type.label_name]


def assert_label_value_exists(metrics: List[MetricData], label: MetricLabelEnum):
    assert find_first_label_value(metrics, label.__class__) == label.label_value



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from frocket.common.config import config


@pytest.fixture(scope="session", autouse=True)
def init_mock_lambda_settings():
    config['lambda.aws.endpoint.url'] = config.get('lambda.aws.endpoint.url', 'http://localhost:9001')
    config['lambda.aws.region'] = config.get('lambda.aws.region', 'us-east-1')
    config['lambda.aws.no.signature'] = 'true'
    config['invoker.lambda.legacy.async'] = 'false'



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections import Counter
from dataclasses import dataclass
from typing import cast, List
import pytest

from frocket.common.config import config
from frocket.common.dataset import DatasetInfo, DatasetPartsInfo, DatasetShortSchema
from frocket.common.tasks.base import BaseTaskResult
from frocket.common.tasks.query import QueryJobResult, PartSelectionMode, AggregationResult, FunnelResult, \
    QueryConditionsResult
from frocket.common.tasks.registration import DatasetValidationMode, RegistrationJobResult
from frocket.datastore.registered_datastores import get_datastore
from frocket.invoker import invoker_api
from frocket.invoker.jobs.query_job import QueryJob
from tests.utils.dataset_utils import TestDatasetInfo, str_and_none_column_values, TestColumn, DEFAULT_ROW_COUNT, \
    DEFAULT_GROUP_COUNT
from tests.utils.task_and_job_utils import build_registration_job, query_job_invoker, registration_job_invoker, \
    build_query_job
from tests.test_query_task import run_query_task, TEST_QUERY_WITH_AGGRS

QUERY_DATASET_NUM_PARTS = 4


@dataclass
class QueryDatasetInfo:
    test_ds: TestDatasetInfo
    local_dataset: DatasetInfo
    remote_dataset: DatasetInfo
    parts_info: DatasetPartsInfo
    short_schema: DatasetShortSchema

    def cleanup(self):
        invoker_api.unregister_dataset(self.local_dataset.id.name)
        invoker_api.unregister_dataset(self.remote_dataset.id.name)
        self.test_ds.cleanup()

    @classmethod
    def shared_instance(cls):
        if not hasattr(cls, '_shared_instance'):
            print("Creating QueryDatasetInfo shared instance (should happen once)")
            test_ds = TestDatasetInfo.build(parts=QUERY_DATASET_NUM_PARTS)
            job = build_registration_job(test_ds.basepath, mode=DatasetValidationMode.SINGLE)
            with registration_job_invoker(job) as invoker:
                job_result = cast(RegistrationJobResult, invoker.run_all_stages())  # Asserts success by default
            local_ds = job_result.dataset

            s3path = test_ds.copy_to_s3()
            job = build_registration_job(s3path, mode=DatasetValidationMode.SINGLE)
            with registration_job_invoker(job) as invoker:
                job_result = cast(RegistrationJobResult, invoker.run_all_stages())
            remote_ds = job_result.dataset

            info = QueryDatasetInfo(test_ds=test_ds,
                                    local_dataset=local_ds,
                                    remote_dataset=remote_ds,
                                    parts_info=get_datastore().dataset_parts_info(local_ds),
                                    short_schema=get_datastore().short_schema(local_ds))
            cls._shared_instance = info
        return cls._shared_instance


@pytest.fixture(scope='module')
def query_dataset_info():
    info = cast(QueryDatasetInfo, QueryDatasetInfo.shared_instance())
    yield
    info.cleanup()


def query_job(query: dict,
              remote_dataset: bool = False,
              worker_can_select_part: bool = None) -> QueryJob:
    query_ds_info = QueryDatasetInfo.shared_instance()
    ds = query_ds_info.remote_dataset if remote_dataset else query_ds_info.local_dataset
    return build_query_job(query, ds, query_ds_info.parts_info, query_ds_info.short_schema, worker_can_select_part)


def run_query(query: dict,
              remote_dataset: bool,
              worker_can_select_part: bool,
              iterations: int = 1,
              expected_groups: int = DEFAULT_GROUP_COUNT * QUERY_DATASET_NUM_PARTS,
              expected_rows: int = DEFAULT_ROW_COUNT * QUERY_DATASET_NUM_PARTS,
              expected_aggregations: List[AggregationResult] = None,
              expected_funnel: FunnelResult = None):

    def sorted_aggregations(aggrs: List[AggregationResult]) -> List[AggregationResult]:
        return sorted(aggrs, key=lambda ag: f"{ag.column}-{ag.type}-{ag.name}")

    for _ in range(iterations):
        job = query_job(query=query, worker_can_select_part=worker_can_select_part, remote_dataset=remote_dataset)
        with query_job_invoker(job) as invoker:
            job_result = cast(QueryJobResult, invoker.run_all_stages())

        if worker_can_select_part:
            assert len(job.dataset_parts_to_publish()) == QUERY_DATASET_NUM_PARTS
        else:
            assert job.dataset_parts_to_publish() is None

        assert job_result.query.matching_groups == expected_groups
        assert job_result.query.matching_group_rows == expected_rows

        if not expected_aggregations:
            assert not job_result.query.aggregations
        else:
            # noinspection PyTypeChecker
            print(expected_aggregations)
            print(job_result.query.aggregations)
            assert sorted_aggregations(expected_aggregations) == sorted_aggregations(job_result.query.aggregations)

        if not expected_funnel:
            assert not job_result.funnel
        else:
            assert job_result.funnel == expected_funnel


def test_empty_query_local_invoker_set():
    # Running a few times just to see that nothing breaks
    run_query(query={}, remote_dataset=False, worker_can_select_part=False, iterations=2)


def test_empty_query_local_worker_set():
    run_query(query={}, remote_dataset=False, worker_can_select_part=True, iterations=2)


def test_empty_query_remote_invoker_set():
    # Running a few times just to see that nothing breaks
    run_query(query={}, remote_dataset=True, worker_can_select_part=False, iterations=2)


def test_empty_query_remote_worker_set():
    run_query(query={}, remote_dataset=True, worker_can_select_part=True, iterations=2)


def test_query_with_aggrs():
    # Get individual task results
    num_parts = QueryDatasetInfo.shared_instance().parts_info.total_parts
    task_results = run_query_task(query=TEST_QUERY_WITH_AGGRS,
                                  test_ds=QueryDatasetInfo.shared_instance().test_ds,
                                  part_selection_mode=PartSelectionMode.SET_BY_INVOKER,
                                  parts=list(range(num_parts)))

    expected_aggr_keys = set()
    for i in range(num_parts):
        expected_aggr_keys.update(str_and_none_column_values(part=i, with_none=False))
    expected_aggr_map = Counter()
    for res in task_results:
        expected_aggr_map.update(res.query_result.query.aggregations[0].value)
    assert sorted(expected_aggr_keys) == sorted(expected_aggr_map.keys())
    expected_count_total = sum(expected_aggr_map.values())

    default_top_count = config.int('aggregations.top.default.count')

    def expected_aggr_results(aggrs_list: List[dict]) -> List[AggregationResult]:
        assert len(aggrs_list) == 2  # Matches test query
        return [
            AggregationResult(**aggrs_list[0],
                              top=default_top_count, value=expected_aggr_map),
            AggregationResult(**aggrs_list[1],
                              top=None, value=expected_count_total),
        ]

    expected_funnel = FunnelResult(
        sequence=[
            QueryConditionsResult(
                matching_groups=DEFAULT_GROUP_COUNT * QUERY_DATASET_NUM_PARTS,
                matching_group_rows=DEFAULT_ROW_COUNT * QUERY_DATASET_NUM_PARTS,
                aggregations=expected_aggr_results(TEST_QUERY_WITH_AGGRS['funnel']['stepAggregations'])
            )
        ],
        end_aggregations=expected_aggr_results(TEST_QUERY_WITH_AGGRS['funnel']['endAggregations'])
    )

    run_query(query=TEST_QUERY_WITH_AGGRS, remote_dataset=False, worker_can_select_part=False,
              expected_aggregations=expected_aggr_results(TEST_QUERY_WITH_AGGRS['query']['aggregations']),
              expected_funnel=expected_funnel)


def test_reduce_ok():
    elem1 = FunnelResult(
        sequence=[
            QueryConditionsResult(matching_groups=100, matching_group_rows=1000, aggregations=[
                AggregationResult(column="c1", type="count", name=None, value=10000, top=None),
                AggregationResult(column="c2", type="countPerValue", name="pv", value={
                    "funny": 20, "few": 4, "not funny": 10, "few2": 3, "a1": 1, "medium1": 5}, top=3)
            ]),
            QueryConditionsResult(matching_groups=50, matching_group_rows=500, aggregations=[
                AggregationResult(column="c1", type="count", name=None, value=20000, top=None),
                AggregationResult(column="c2", type="countPerValue", name="pv", value={
                    "funny": 40, "few": 8, "not funny": 20, "few2": 6, "medium": 2, "another": 1}, top=6)
            ]),
        ],
        end_aggregations=[
            AggregationResult(column="ce2", type="countPerValue", name="pv", top=2,
                              value={f"value{i}": i*10 for i in range(1, 101)})
        ]
    )

    elem2 = FunnelResult(
        sequence=[
            QueryConditionsResult(matching_groups=1000, matching_group_rows=10000, aggregations=[
                AggregationResult(column="c1", type="count", name=None, value=100000, top=None),
                AggregationResult(column="c2", type="countPerValue", name="pv", value={
                    "funny": 200, "few": 40, "not funny": 100, "few2": 30, "a2": 10, "medium2": 499}, top=3)
            ]),
            QueryConditionsResult(matching_groups=500, matching_group_rows=5000, aggregations=[
                AggregationResult(column="c1", type="count", name=None, value=200000, top=None),
                AggregationResult(column="c2", type="countPerValue", name="pv", value={
                    "super-funny": 4001, "few2": 1}, top=6)
            ]),
        ],
        end_aggregations=[
            AggregationResult(column="ce2", type="countPerValue", name="pv", top=2,
                              value={f"value{i}": i * 3 for i in range(1, 201)})
        ]
    )

    elems = [elem1]*100 + [elem2]

    expected_result = FunnelResult(
            sequence=[
                QueryConditionsResult(matching_groups=11000, matching_group_rows=110000, aggregations=[
                    AggregationResult(column="c1", type="count", name=None, value=1100000, top=None),
                    AggregationResult(column="c2", type="countPerValue", name="pv", value={
                        "funny": 2200, "not funny": 1100, "medium1": 500}, top=3)
                ]),
                QueryConditionsResult(matching_groups=5500, matching_group_rows=55000, aggregations=[
                    AggregationResult(column="c1", type="count", name=None, value=2200000, top=None),
                    AggregationResult(column="c2", type="countPerValue", name="pv", value={
                        "super-funny": 4001, "funny": 4000, "not funny": 2000, "few": 800, "few2": 601, "medium": 200},
                                      top=6)
                ]),
            ],
            end_aggregations=[
                AggregationResult(column="ce2", type="countPerValue", name="pv", top=2,
                                  value={"value100": 100300, "value99": 99297})
            ]
        )

    reduced_result = cast(FunnelResult, FunnelResult.reduce(elems))
    assert reduced_result == expected_result


def test_reduce_with_none_empty():
    elems = [
        QueryConditionsResult(matching_groups=0, matching_group_rows=0, aggregations=None),
        QueryConditionsResult(matching_groups=1, matching_group_rows=1, aggregations=None),
        QueryConditionsResult(matching_groups=2, matching_group_rows=2, aggregations=None)
    ]
    res = cast(QueryConditionsResult, QueryConditionsResult.reduce(elems))
    assert res.matching_groups == 3 and res.matching_group_rows == 3
    assert res.aggregations is None

    elems = [
        QueryConditionsResult(matching_groups=0, matching_group_rows=0, aggregations=[]),
        QueryConditionsResult(matching_groups=1, matching_group_rows=1, aggregations=[]),
        QueryConditionsResult(matching_groups=2, matching_group_rows=2, aggregations=[])
    ]
    res = cast(QueryConditionsResult, QueryConditionsResult.reduce(elems))
    assert res.matching_groups == 3 and res.matching_group_rows == 3
    assert res.aggregations == []


@pytest.mark.xfail(strict=True)
def test_reduce_partially_none_xfail():
    elems = [
        QueryConditionsResult(matching_groups=0, matching_group_rows=0, aggregations=None),
        QueryConditionsResult(matching_groups=1, matching_group_rows=1, aggregations=None),
        QueryConditionsResult(matching_groups=2, matching_group_rows=2, aggregations=[])
    ]
    QueryConditionsResult.reduce(elems)


@pytest.mark.xfail(strict=True)
def test_reduce_wrong_class_xfail():
    elems = [
        QueryConditionsResult(matching_groups=0, matching_group_rows=0, aggregations=None),
        QueryConditionsResult(matching_groups=1, matching_group_rows=1, aggregations=None),
    ]
    FunnelResult.reduce(elems)


@pytest.mark.xfail(strict=True)
def test_reduce_aggrs_len_xfail():
    elems = [
        QueryConditionsResult(matching_groups=0, matching_group_rows=0, aggregations=[
            AggregationResult(column="c", type="count", name="n", value=1, top=None)
        ]),
        QueryConditionsResult(matching_groups=1, matching_group_rows=1, aggregations=[
            AggregationResult(column="c", type="count", name="n", value=1, top=None),
            AggregationResult(column="c2", type="count", name="n2", value=1, top=None)
        ]),
    ]
    QueryConditionsResult.reduce(elems)


@pytest.mark.xfail(strict=True)
def test_reduce_different_aggrs_xfail():
    elems = [
        QueryConditionsResult(matching_groups=0, matching_group_rows=0, aggregations=[
            AggregationResult(column="c1", type="count", name="n", value=1, top=None)
        ]),
        QueryConditionsResult(matching_groups=1, matching_group_rows=1, aggregations=[
            AggregationResult(column="c2", type="count", name="n", value=1, top=None),
        ]),
    ]
    QueryConditionsResult.reduce(elems)


@pytest.mark.xfail(strict=True)
def test_reduce_different_value_types_xfail():
    elems = [
        QueryConditionsResult(matching_groups=0, matching_group_rows=0, aggregations=[
            AggregationResult(column="c1", type="count", name="n", value=1, top=None)
        ]),
        QueryConditionsResult(matching_groups=1, matching_group_rows=1, aggregations=[
            AggregationResult(column="c1", type="count", name="n", value={"a": 1}, top=None),
        ]),
    ]
    QueryConditionsResult.reduce(elems)


# TODO explain this mess...
def test_build_retry_task():
    num_parts = QueryDatasetInfo.shared_instance().parts_info.total_parts
    job = query_job(
        query={'query': {'aggregations': [{'column': TestColumn.str_userid.value, 'type': 'count'}]}},
        worker_can_select_part=True)
    iterations = 5
    for _ in range(iterations):
        with query_job_invoker(job) as invoker:
            invoker.prerun()
            invoker.enqueue_tasks()
            results = invoker.run_tasks()
            assert len(results) == num_parts

            per_task_retries = 2
            for _ in range(per_task_retries):
                results = invoker.retry(task_index=0)
                results = invoker.retry(task_index=num_parts - 1)

            assert len(results) == num_parts + (per_task_retries * 2)
            results = sorted(results, key=lambda res: res.task_index)

            def ensure_results_equal(results: List[BaseTaskResult]):
                dicts = []
                for res in results:
                    d = res.to_dict()
                    d.pop('metrics')
                    dicts.append(d)
                assert all([d == dicts[0] for d in dicts[1:]])

            ensure_results_equal(results[:(per_task_retries+1)])
            ensure_results_equal(results[-(per_task_retries+1):])

            # Now ensure that only one result per task index (== part idx in this case) is used for aggregation,
            # regardless of no. of retries
            invoker.complete()
            job_result = cast(QueryJobResult, invoker.build_result())
            assert job_result.query.matching_groups == DEFAULT_GROUP_COUNT * QUERY_DATASET_NUM_PARTS
            assert job_result.query.matching_group_rows == DEFAULT_ROW_COUNT * QUERY_DATASET_NUM_PARTS
            assert job_result.query.aggregations[0].value == DEFAULT_ROW_COUNT * QUERY_DATASET_NUM_PARTS



"""
Invoke tasks by invoking an AWS Lambda function asynchronously.

This is a great feature of Lamdba, which implictly manages a queue of invocation for you with configurable retention
(probably based on SQS). As long as the concurrent invocations limit in your account/burst limit of the AWS region are
not reached, AWS will launch Lambdas for queued invocation immediately, with no meaningful delay. This also prevents
getting rate-limited on momentary invocation spikes.

A few important notes:

1. As noted in the setup guide, the retry count for the Lambda function *should be set to zero*, as it's the invoker's
job to launch retries with slightly different arguments, based on its own configuration, with logic that is agnostic
to whether the actual invoker is using Lambdas or anything else (which does not have its optional retry feature).

2. Unfortunately, there's no API for batch Lambda invocation, so we're invoking one by one with multiple threads -
and still the time to invoke all tasks can add up to 1-2 seconds or more.
TODO backlog optimize! this also hurts caching as not all tasks get their fair chance to pick a locally cached part.

3. The InvokeAsync() Lambda API is considered deprecated and replaced by the 'InvocationType' parameter in Invoke().
However, the InvokeAsync API currently seems to take about half the time to return! Which one to use is configurable.

TODO backlog stress-test queue limits till reaching rate limiting (status 429).
TODO backlog for each invocation, add its actual invoke time as parameter
 (now we only measure time since invocation of all tasks started)
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import time
import concurrent.futures
from typing import cast
import boto3
from botocore.client import BaseClient
from botocore.config import Config
from frocket.common.serializable import Envelope
from frocket.common.tasks.base import BaseTaskRequest, BaseApiResult
from frocket.invoker.impl.async_invoker import AsyncInvoker
from frocket.common.config import config

logger = logging.getLogger(__name__)

DEBUG_PRINT_PAYLOADS = config.bool("invoker.lambda.debug.payload")
LAMBDA_ASYNC_OK_STATUS = 202


def _worker_task(req: BaseTaskRequest, client: BaseClient, lambda_name: str) -> BaseApiResult:
    """Run by the thread pool below."""
    # noinspection PyBroadException
    try:
        result = None
        json_payload = Envelope.seal_to_json(req)  # Encodes the actual object and its type, for correct decoding later.
        if DEBUG_PRINT_PAYLOADS:
            logger.debug(json_payload)

        legacy_invoke_async = config.bool("invoker.lambda.legacy.async")
        status_field = 'Status' if legacy_invoke_async else 'StatusCode'

        if legacy_invoke_async:
            response = client.invoke_async(FunctionName=lambda_name, InvokeArgs=json_payload)
        else:
            response = client.invoke(FunctionName=lambda_name, InvocationType='Event', Payload=json_payload)

        if response[status_field] == LAMBDA_ASYNC_OK_STATUS:
            result = BaseApiResult(success=True, error_message=None)
        else:
            message = f"Response status differs from expected ({LAMBDA_ASYNC_OK_STATUS}): {response}"
            result = BaseApiResult(success=False, error_message=message)
    except Exception as e:
        result = BaseApiResult(success=False, error_message=f"Failed to invoke lambda function '{lambda_name}': {e}")
    return result


class AwsLambdaInvoker(AsyncInvoker):
    def _enqueue(self, requests) -> None:
        lambda_name = config.get('invoker.lambda.name')
        num_threads = config.int('invoker.lambda.threads')
        boto_config = Config(**config.aws_config_dict(service='lambda'))
        client = boto3.client('lambda',
                              **config.aws_client_settings(service='lambda'),
                              config=boto_config)
        logger.debug(f"Invoking lambdas, name: {lambda_name}, no. of invocations: {len(requests)}"
                     f", no. of invoker threads: {num_threads}")
        futures = []
        start_invoke_time = time.time()
        # TODO backlog consider lifecycle of the thread pool
        with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor:
            for req in requests:
                futures.append(executor.submit(_worker_task, req, client, lambda_name))
            futures = concurrent.futures.as_completed(futures)  # Wait till all complete!
            executor.shutdown()

        error_message = None
        for f in futures:
            assert f.done()
            if f.cancelled():
                error_message = "Lambda invocation interrupted"
            elif f.exception():
                error_message = f"Invocation failed with error: {f.exception()}"
            else:
                result = f.result()
                if not result or type(result) is not BaseApiResult:
                    error_message = f"Invocation returned with response: {result}"
                result = cast(BaseApiResult, result)
                if not result.success:
                    error_message = result.error_message
            if error_message:
                break

        if error_message:
            logger.error(error_message)
            raise Exception(error_message)
        else:
            logger.info(f"Async invocation done in {time.time() - start_invoke_time:.3f}")



"""Base classes for registered datasets and their metadata."""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from enum import auto
from datetime import datetime, timezone
from typing import Optional, List, Dict
from dataclasses import dataclass, field
from frocket.common.serializable import SerializableDataClass, AutoNamedEnum

logger = logging.getLogger(__name__)


class PartNamingMethod(AutoNamedEnum):
    """
    For future use: currently the full list of dataset filenames is stored as metadata, however if a consistent
    numbering pattern should be identified, it may be useful both for more compact metadata and for a more predictable
    part index -> filename mapping.
    """
    RUNNING_NUMBER = auto()
    LIST = auto()


@dataclass(frozen=True)
class DatasetId(SerializableDataClass):
    """
    The main reason why this class exists: datasets can be re-registered multiple times with the same name, but any
    caching behavior should be sensitive to the registered date and become invalid on re-registration.
    In concrete terms, caching should be based on DatasetId keys (which are immutable) rather than a dataset name.

    Re-registering a dataset is useful, in cases such as:
    1. When you don't need to manage revisions yourself (via specifying a new dataset name and un-registering old ones).
    2. As an alias to the current version (datasets are only metadata, you can register the same physical files N times)
    3. If the datafiles were found to be incomplete/invalid, and after fixing the issue you want to invalidate caching.
    """
    name: str
    registered_at: datetime

    @classmethod
    def now(cls, name: str):
        return DatasetId(name, registered_at=datetime.now(tz=timezone.utc))


@dataclass(frozen=True)
class DatasetPartId(SerializableDataClass):
    """Specifies a single part (file) in a dataset version (see documetation for DatasetId above!)."""
    dataset_id: DatasetId
    path: str
    part_idx: int


@dataclass(frozen=True)
class DatasetInfo(SerializableDataClass):
    """
    Basic metadata for a dataset.

    This class should be kept pretty small, as it's passed along in task requests.
    More detailed metadata is found in the data schema object, which is stored separately and read when needed
    (and also exists in both short and full versions)
    """
    basepath: str
    total_parts: int
    id: DatasetId
    group_id_column: str  # The column by which the dataset is partitioned, and grouping is done.
    timestamp_column: str  # The column by which timeframe conditions and funnels are run.


@dataclass(frozen=True)
class DatasetPartsInfo(SerializableDataClass):
    """Holds the list of files in the dataset. Separate from DatasetInfo only due to size (this data is usually not
    needed to be sent in network calls)."""
    naming_method: PartNamingMethod
    total_parts: int
    total_size: int
    running_number_pattern: Optional[str] = field(default=None)
    filenames: Optional[List[str]] = field(default=None)

    def fullpaths(self, parent: DatasetInfo) -> List[str]:
        parentpath = parent.basepath if parent.basepath.endswith('/') else parent.basepath + '/'

        if self.naming_method == PartNamingMethod.LIST:
            assert (self.filenames and len(self.filenames) == parent.total_parts)
            return [parentpath + filename for filename in self.filenames]
        else:
            assert self.running_number_pattern
            return [parentpath + self.running_number_pattern.format(idx)
                    for idx in range(parent.total_parts)]


class DatasetColumnType(AutoNamedEnum):
    INT = auto()
    FLOAT = auto()
    BOOL = auto()
    # Categorical columns are not a separate type to the query engine. That designation exists and is used separately.
    STRING = auto()


@dataclass(frozen=True)
class DatasetColumnAttributes(SerializableDataClass):
    """
    The 'full' information on each column. TODO backlog use polymorphism? (needs support in de-serialization)

    For columns which were either saved by Pandas as categoricals, or are identified during registration to be such,
    store a mapping of top N values (configurable) to their their normalized share in the dataset. Since registration
    does not read all files but only a sample, that ratio cannot be an absolute number or the exact ratio - but still
    useful for clients.

    cat_unique_ratio is the ratio of unique value count to all values (or: series.nunique()/len(series)), and may be
    a useful rough indicator to how much RAM is saved (and str.match() operations sped-up!) by the categorical
    representation. Columns are determined to be loaded as categorical if this value is lower than configured.
    Loading of columns as categoricals is also usually much faster, but that greatly depends on whether a dictionary
    was saved for that column in the Parquet file or not - so it depends on the tool used to create these files.
    """
    numeric_min: Optional[float] = None
    numeric_max: Optional[float] = None
    categorical: bool = False
    cat_top_values: Optional[Dict[str, float]] = None
    cat_unique_ratio: Optional[float] = None


@dataclass(frozen=True)
class DatasetColumn(SerializableDataClass):
    name: str
    dtype_name: str
    coltype: DatasetColumnType
    colattrs: DatasetColumnAttributes


@dataclass(frozen=True)
class DatasetShortSchema(SerializableDataClass):
    """Schema, the short version - typically all you may need."""
    columns: Dict[str, DatasetColumnType]
    min_timestamp: float
    max_timestamp: float
    # In files created by Pandas with its metadata intact in the Parquet file, columns marked as categoricals.
    source_categoricals: List[str] = field(default=None)
    # Columns detected during registration to be good candidates for explicitly loading as categoricals (by PyArrow).
    potential_categoricals: List[str] = field(default=None)


@dataclass(frozen=True)
class DatasetSchema(SerializableDataClass):
    group_id_column: str
    timestamp_column: str
    columns: Dict[str, DatasetColumn]
    # Just the names->dtypes of all columns not (currently) supported.
    unsupported_columns: Dict[str, str]

    def short(self) -> DatasetShortSchema:
        """Make short from full."""
        cols = {name: col.coltype for name, col in self.columns.items()}
        source_categoricals = []
        potential_categoricals = []
        for name, col in self.columns.items():
            if col.colattrs.categorical:
                if col.dtype_name == 'category':
                    source_categoricals.append(name)
                else:
                    potential_categoricals.append(name)
        ts_attrs = self.columns[self.timestamp_column].colattrs
        min_ts = ts_attrs.numeric_min
        max_ts = ts_attrs.numeric_max

        return DatasetShortSchema(columns=cols,
                                  source_categoricals=source_categoricals,
                                  potential_categoricals=potential_categoricals,
                                  min_timestamp=min_ts, max_timestamp=max_ts)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import abstractmethod, ABCMeta
from typing import List, Optional, Set
from frocket.common.dataset import DatasetPartId, DatasetPartsInfo
from frocket.common.metrics import LabelsDict
from frocket.common.tasks.base import BaseTaskRequest, BaseTaskResult, BaseJobResult, JobStatus, ErrorMessage
from frocket.common.tasks.async_tracker import AsyncJobStatusUpdater


class Job(metaclass=ABCMeta):
    """
    For each job type (registratioThen, qeury, and future ones) there is a concrete subclass.
    That concrete class is handed to the invoker object, which is agnostic to the job details but calls the
    job's method in a set order.

    The flow, at high level:
    1. On prerun(), the job validates its arguments (and can fail by returnning an error message) and can prepare data
       for building tasks.

    2. When build_tasks() is called by the invoker - return a list of concrete task request objects,
       all with attempt no. 0.

    3. If the job supports task self-selection by workers, it should override dataset_parts_to_publish() and
       return a list of parts to be consumed by workers (workers would try to select parts they have cached locally).
       This list is published by the data store before tasks are invoked.

    4.  In case the invoker decides to retry a task, it calls build_retry_task() to create a specific retry task

    5. After all tasks have completed, either successfully or not, complete() is called to run any validations on the
       final results of all tasks, and perform any needed aggregations. The job may fail at this stage if the results of
       tasks, taken together, are invalid.

    6. Lastly, build_result() is called to construct the final job result.
       At this stage, the final success status of the job should not change.
    """
    _request_id = None
    _labels = {}

    @property
    def request_id(self) -> Optional[str]:
        return self._request_id

    @request_id.setter
    def request_id(self, request_id: str):
        self._request_id = request_id

    def prerun(self, async_updater: AsyncJobStatusUpdater = None) -> Optional[ErrorMessage]:
        pass

    @abstractmethod
    def build_tasks(self) -> List[BaseTaskRequest]:
        pass

    def dataset_parts_to_publish(self) -> Optional[Set[DatasetPartId]]:
        return None

    @abstractmethod
    def total_tasks(self) -> int:
        pass

    @abstractmethod
    def build_retry_task(self, attempt_no: int, task_index: int) -> BaseTaskRequest:
        pass

    def complete(self,
                 tasks_final_status: JobStatus,
                 latest_task_results: List[BaseTaskResult],
                 async_updater: AsyncJobStatusUpdater = None) -> JobStatus:
        return tasks_final_status

    @abstractmethod
    def build_result(self,
                     base_attributes: dict,
                     final_status: JobStatus,
                     latest_task_results: List[BaseTaskResult]) -> BaseJobResult:
        pass

    @property
    def metric_labels(self) -> LabelsDict:
        return self._labels

    @abstractmethod
    def parts_info(self) -> Optional[DatasetPartsInfo]:
        pass



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Iterable
import pandas as pd
import numpy as np


def filter_by_isin(df: pd.DataFrame, column: str, values: Iterable) -> pd.DataFrame:
    """
    For the given DataFrame, return only rows where df[column] is in the given values.
    This is a surprisingly faster alternative to built-in Pandas/NumPy functions: df[np.isin(df[column], values)]
    A value can appear in multiple rows (e.g. the same user ID appearing multiple rows)

    TODO Merge a [Numba-based isin()](https://stackoverflow.com/questions/53046473/numpy-isin-performance-improvement)
     function, compiled AOT for relevant array dtypes. This would be arch-dependent and optional (with fallback)
    """
    # First, create a "map" series from all possible values in the column => whether they should pass the filter
    all_ids = df[column].unique()
    is_id_relevant = pd.Series(np.zeros(len(all_ids)), index=all_ids).astype('bool')  # Default false
    is_id_relevant.loc[values] = True

    # Create a boolean mask for column, based on the mapping above. Grab the raw array.
    mask = is_id_relevant[df[column]].values
    # Apply mask
    return df[mask]


def add_column_by_value_map(df: pd.DataFrame, keys_column: str, values_map_series: pd.Series, new_column: str) -> None:
    """
    Add a new column to the given df. For each row, df[new_column] will be set to an appropriate value from
    values_map_series: the value whose index is df[keys_column] in that row.

    e.g. given a DF of user activities having a userId column (with potentially multiple rows per user), and a
    values_map_series whose unique index is a User ID, and its values are the age of that user, the function will add
    a new column to the given DF with the age of that row's user ID

    If a value in keys_column does not have a matching index in values_map_series, the cell value would be NaN.
    This function is optimized for performance.

    The given DF is modified inplace.
    """
    # Create a new mapping between ALL unique values of IDs of df[keys_column] and their matching value (or NaN)
    unique_keys = df[keys_column].unique()
    key_to_value = pd.Series(data=np.nan, index=unique_keys)
    key_to_value.loc[values_map_series.index] = values_map_series

    # Now we can create the new column, using the mapping
    df[new_column] = key_to_value[df[keys_column]].values



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import time
from collections import Counter
import numpy as np
from typing import List, Optional, cast, Dict
from frocket.common.config import config
from frocket.common.dataset import DatasetInfo, DatasetPartId, DatasetPartsInfo, \
    DatasetSchema, DatasetColumnType, DatasetColumnAttributes, DatasetColumn, DatasetId
from frocket.common.helpers.storage import storage_handler_for
from frocket.common.metrics import JobTypeLabel, DATASET_LABEL
from frocket.common.tasks.base import ErrorMessage
from frocket.common.tasks.registration import RegistrationTaskRequest, RegistrationJobResult, \
    RegistrationTaskResult, DatasetValidationMode, RegisterArgs
from frocket.common.helpers.utils import sample_from_range, bytes_to_ndarray
from frocket.datastore.registered_datastores import get_blobstore, get_datastore
from frocket.invoker.jobs.job import Job

logger = logging.getLogger(__name__)

VALIDATION_SAMPLE_RATIO = config.float('validation.sample.ratio')  # Ratio of sample size to dataset total parts
VALIDATION_MAX_SAMPLES = config.int('validation.sample.max')  # Hard limit on on. of samples
# How many 'top N' values to store in schema for categoricl fields
CATEGORICAL_TOP_COUNT = config.int('dataset.categorical.top.count')
# To be included in Top N list, a value should have at least this ratio of occurences to total row count
CATEGORICAL_TOP_MIN_PCT = config.float('dataset.categorical.top.minpct')


class RegistrationJob(Job):
    def __init__(self, args: RegisterArgs):
        self._args = args
        self._dataset = None
        self._parts_info = None
        self._validate_uniques = None
        self._sampled_parts = None
        self._final_schema = None

    def prerun(self, async_updater=None):
        # Before validation tasks are run, finc out which parts (files) there are in the given path
        if async_updater:
            async_updater.update(message=f"Discovering files in {self._args.basepath}")

        # TODO backlog run file discovery in a worker task, not directly by the invoker, for better separation of
        #  concerns and *permissions* (invoker won't need list permission to remote storage)
        parts_info = self._discover_parts(self._args.basepath, self._args.pattern)
        if parts_info.total_parts == 0:
            return f"No files found at {self._args.basepath}"

        # If needed, adjust validation mode and whether to validate uniques by actual no. of parts
        validate_uniques = self._args.validate_uniques
        if parts_info.total_parts == 1:
            mode = DatasetValidationMode.SINGLE
            validate_uniques = False
        elif parts_info.total_parts == 2 and self._args.validation_mode == DatasetValidationMode.SAMPLE:
            mode = DatasetValidationMode.FIRST_LAST
        else:
            mode = self._args.validation_mode

        # The dataset will be persisted to datastore once registration completes successfully
        dataset = DatasetInfo(id=DatasetId.now(self._args.name),
                              basepath=self._args.basepath,
                              total_parts=parts_info.total_parts,
                              group_id_column=self._args.group_id_column,
                              timestamp_column=self._args.timestamp_column)

        self._sampled_parts = self._choose_sampled_parts(dataset, parts_info, mode)
        self._dataset = dataset
        self._parts_info = parts_info
        self._validate_uniques = validate_uniques
        self._labels = {
            JobTypeLabel.REGISTER.label_name: JobTypeLabel.REGISTER.label_value,
            DATASET_LABEL: dataset.id.name
        }

    @staticmethod
    def _discover_parts(basepath: str, pattern: str):
        handler = storage_handler_for(basepath)  # Returns the appropriate storage handler based on the path format
        result = handler.discover_files(pattern)
        if result.total_parts > 0:
            logger.info(f"Number of part files: {result.total_parts}, "
                        f"total size {result.total_size / (1024 ** 2):.2f}mb")
            if result.filenames:
                logger.debug(f"List of files: {result.filenames}")
        return result

    @staticmethod
    def _choose_sampled_parts(dataset: DatasetInfo,
                              parts: DatasetPartsInfo,
                              mode: DatasetValidationMode) -> List[DatasetPartId]:
        """Determine the list of parts to sample, and return as DatasetPartIds"""
        if mode == DatasetValidationMode.SINGLE:
            chosen_indexes = [0]
        elif mode == DatasetValidationMode.FIRST_LAST:
            chosen_indexes = [0, parts.total_parts - 1]
        else:
            # In SAMPLE mode, always include first & last part, plus zero..VALIDATION_MAX_SAMPLES extra parts
            preselected = [0, parts.total_parts - 1]
            chosen_indexes = sample_from_range(range_max=parts.total_parts,
                                               sample_ratio=VALIDATION_SAMPLE_RATIO,
                                               max_samples=VALIDATION_MAX_SAMPLES,
                                               preselected=preselected)

        all_paths = parts.fullpaths(dataset)
        result = [DatasetPartId(dataset_id=dataset.id, path=all_paths[part_index], part_idx=part_index)
                  for part_index in chosen_indexes]
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Parts chosen to be sampled: {result}")
        return result

    def parts_info(self):
        return self._parts_info

    def total_tasks(self):
        return len(self._sampled_parts)

    def build_tasks(self):
        requests = [self._build_task(task_index=i) for i in range(self.total_tasks())]
        return requests

    def build_retry_task(self, attempt_no, task_index):
        return self._build_task(task_index, attempt_no=attempt_no)

    def _build_task(self, task_index: int, attempt_no: int = 0) -> RegistrationTaskRequest:
        request = RegistrationTaskRequest(
            request_id=self._request_id,
            invoke_time=time.time(),
            invoker_set_task_index=task_index,  # Task index & part are always set by invoker in this job
            attempt_no=attempt_no,
            dataset=self._dataset,
            part_id=self._sampled_parts[task_index],
            return_group_ids=self._validate_uniques)
        return request

    def complete(self, job_status, latest_task_results, async_updater=None):
        """Now that all individual tasks have completed, check for consistency between task results, and build the
        aggregated schema."""
        if not job_status.success:
            return job_status

        try:
            # Validation across task results
            task_results = cast(List[RegistrationTaskResult], latest_task_results)
            if self._validate_uniques:
                message = self._check_uniques(task_results)
                if message:
                    return job_status.with_error(message=message)

            message = self._compare_columns(task_results)
            if message:
                return job_status.with_error(message=message)

            # All valid, now aggregate the final schema
            self._final_schema = self._aggregate_schema(task_results)
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Short schema:" + self._final_schema.short().to_json(indent=2))

            # Register to datastore
            datastore = get_datastore()
            datastore.write_dataset_info(self._dataset, self._parts_info, self._final_schema)
            logger.info(f"Dataset '{self._dataset.id.name}' successfully registered in {datastore}")
            return job_status
        except Exception as e:
            logger.exception("Validation of sampled parts failed unexpectedly")
            return job_status.with_error(message=str(e))

    def _aggregate_schema(self, task_results: List[RegistrationTaskResult]) -> DatasetSchema:
        full_schemas = [res.dataset_schema for res in task_results]
        new_columns: Dict[str, DatasetColumn] = {}
        for name, col in full_schemas[0].columns.items():
            all_colattrs = [s.columns[name].colattrs for s in full_schemas]

            if col.coltype == DatasetColumnType.BOOL:
                new_attrs = col.colattrs  # No change

            elif col.coltype in [DatasetColumnType.INT, DatasetColumnType.FLOAT]:
                new_attrs = DatasetColumnAttributes(
                    numeric_min=min([at.numeric_min for at in all_colattrs]),
                    numeric_max=max([at.numeric_max for at in all_colattrs]))

            elif col.coltype == DatasetColumnType.STRING:
                is_categorical = col.colattrs.categorical
                if is_categorical and False in [at.categorical for at in all_colattrs]:
                    logger.info(f"Not all part schemas have column {name} as categorical")
                    is_categorical = False

                if is_categorical:
                    approx_uniques_ratio = sum([at.cat_unique_ratio for at in all_colattrs]) / len(all_colattrs)
                    approx_top_values = self._categorical_top_values(all_colattrs)
                else:
                    approx_uniques_ratio = None
                    approx_top_values = None

                new_attrs = DatasetColumnAttributes(categorical=is_categorical,
                                                    cat_unique_ratio=approx_uniques_ratio,
                                                    cat_top_values=approx_top_values)

            else:
                raise Exception(f"Column {name} has unkonwn type {col.coltype}")
            assert new_attrs
            new_columns[name] = DatasetColumn(name=name, dtype_name=col.dtype_name,
                                              coltype=col.coltype, colattrs=new_attrs)

        # TODO backlog consider merging list of unsupported columns from all sampled parts
        return DatasetSchema(columns=new_columns,
                             unsupported_columns=full_schemas[0].unsupported_columns,
                             group_id_column=full_schemas[0].group_id_column,
                             timestamp_column=full_schemas[0].timestamp_column)

    @staticmethod
    def _categorical_top_values(all_colattrs):
        all_tops = [at.cat_top_values for at in all_colattrs if at.cat_top_values]
        if len(all_tops) == 0:
            return None
        aggr_tops = Counter()
        for top in all_tops:
            aggr_tops.update(top)
        highest_aggr = aggr_tops.most_common(CATEGORICAL_TOP_COUNT)
        highest_normalized = [(name, value / len(all_colattrs))
                              for (name, value) in highest_aggr]
        approx_top_values = {name: value for name, value in highest_normalized
                             if value >= CATEGORICAL_TOP_MIN_PCT}
        return approx_top_values

    def _compare_columns(self, task_results: List[RegistrationTaskResult]) -> Optional[ErrorMessage]:
        short_schemas = [res.dataset_schema.short() for res in task_results]
        first_schema = short_schemas[0]

        for i, curr_schema in enumerate(short_schemas):
            if i > 0 and first_schema.columns != curr_schema.columns:
                message = f"Schemas differ!\n" \
                          f"In {self._sampled_parts[task_results[0].task_index].path} " \
                          f"({len(first_schema.columns)} columns):\n{first_schema} vs.\n" \
                          f"In {self._sampled_parts[task_results[i].task_index].path} " \
                          f"({len(curr_schema.columns)} columns):\n{curr_schema}"
                return message
        return None

    @staticmethod
    def _check_uniques(task_results: List[RegistrationTaskResult]) -> Optional[ErrorMessage]:
        # This part can take a bit of memory,
        # so at least we don't use list comprehension to read all blobs before processing any.
        id_arrays = []
        for res in task_results:
            blob = get_blobstore().read_blob(res.group_ids_blob_id)
            if not blob:
                raise Exception(f"Blob not found, ID: {res.group_ids_blob_id}")
            id_array = bytes_to_ndarray(blob)
            id_arrays.append(id_array)
            get_blobstore().delete_blob(res.group_ids_blob_id)

        all_ids = np.concatenate(id_arrays)
        all_unique = np.unique(all_ids)
        if len(all_ids) == len(all_unique):
            return None
        else:
            message = f"Group IDs are not unique across the {len(id_arrays)} sampled files! " \
                      f"Count of unique IDs merged from all sampled files ({len(all_unique)}) " \
                      f"should equal the simple sum of the unique IDs counts from each file ({len(all_ids)})"
            return message

    def build_result(self, base_attributes, final_status, latest_task_results):
        return RegistrationJobResult(
            **base_attributes,
            dataset=self._dataset if final_status.success else None)

    @property
    def sampled_parts(self) -> Optional[List[DatasetPartId]]:
        return self._sampled_parts

    @property
    def dataset(self) -> Optional[DatasetInfo]:
        return self._dataset

    @property
    def parts(self) -> Optional[DatasetPartsInfo]:
        return self._parts_info



"""
While the query schema is generally JSON-based (good for machines) rather then textual (like SQL,
supposedly human-friendly or at least more concise), there's one exception: an optional 'relation' expression allowing
to specify arbitrarily complex and/or relations between conditions, rather than just and/or over all.

The RelationParser class validates and breaks down the expression to a list of elements. However, it does not transform
them back into a Pandas query or similar - that is the query engine's responsibility and may change independently.

Note that conditions may be represented either by index ($0, $3, etc.) or by name - for named conditions.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import Type, List, Optional
from parsimonious.grammar import Grammar, NodeVisitor
from parsimonious.nodes import Node
from dataclasses import dataclass
from parsimonious.exceptions import ParseError, VisitationError
from abc import ABCMeta
from frocket.common.validation.consts import RELATION_OPS, map_condition_names, CONDITION_COLUMN_PREFIX
from frocket.common.validation.path_visitor import PathVisitor
from frocket.common.validation.error import ValidationErrorKind, QueryValidationError
from frocket.common.tasks.base import ErrorMessage

logger = logging.getLogger(__name__)

# TODO backlog fix the grammar to require whitespace between conditions and wordy-operators (and,or),
#  but not around symbol ops (&&, ||)
# TODO backlog fix "DeprecationWarning: invalid escape sequence \$"
RELATION_EXPRESSION_GRAMMAR = Grammar(
    """
    expression = (identifier / (open_paren ws? expression ws? close_paren)) (ws? op ws? expression)*
    identifier = condition_name / condition_id
    condition_name = ~r"\$[A-Z][A-Z_0-9]*"i
    condition_id = ~r"\$[0-9]+"
    op = "and" / "or" / "&&" / "||"
    ws = ~r"\s*"
    open_paren  = "("
    close_paren = ")"
    """)


@dataclass(frozen=True)
class RelationParserContext:
    condition_count: int
    named_conditions: dict
    column_prefix: str


@dataclass
class RBaseElement(metaclass=ABCMeta):
    text: str
    ctx: RelationParserContext
    condition_id: Optional[int] = None

    def validate(self) -> Optional[ErrorMessage]:
        pass

    def __str__(self):
        return f"{self.__class__.__name__}('{self.text}')"


@dataclass
class RTextElement(RBaseElement):
    pass


@dataclass
class RConditionBaseElement(RBaseElement):
    pass


@dataclass
class RConditionId(RConditionBaseElement):
    def validate(self):
        cid = int(self.text[1:])
        if cid >= self.ctx.condition_count:
            return f"Condition no. {cid} does not exist"
        self.condition_id = cid


@dataclass
class RConditionName(RConditionBaseElement):
    def validate(self):
        cname = self.text[1:]
        cid = self.ctx.named_conditions.get(cname, None)
        if cid is not None:  # Can be zero
            self.condition_id = cid
        else:
            return f"Condition named {self.text[1:]} does not exist"


@dataclass
class ROperator(RBaseElement):
    def validate(self):
        if self.text not in RELATION_OPS:
            return f"Operator {self.text} not in {RELATION_OPS}"


# noinspection PyMethodMayBeStatic,PyUnusedLocal
@dataclass
class RelationExpressionVisitor(NodeVisitor):
    """
    Used by the RelationParser to build the element list.
    Note that while the grammar is hierarchical, the resulting list isn't (no need, currently).
    """
    ctx: RelationParserContext

    def _build_element(self, node: Node, cls: Type[RBaseElement]):
        # noinspection PyArgumentList
        return cls(node.text, self.ctx)

    def visit_ws(self, node: Node, visited_children):
        return None  # Ignore whitespaces

    def visit_op(self, node: Node, visited_children):
        return self._build_element(node, ROperator)

    def visit_open_paren(self, node: Node, visited_children):
        return self._build_element(node, RTextElement)

    def visit_close_paren(self, node: Node, visited_children):
        return self._build_element(node, RTextElement)

    def visit_identifier(self, node: Node, visited_children):
        """Return the actual condition name / ID element (see grammar: identifier wraps conditions)."""
        return visited_children[0]

    def visit_condition_name(self, node: Node, visited_children):
        return self._build_element(node, RConditionName)

    def visit_condition_id(self, node: Node, visited_children):
        return self._build_element(node, RConditionId)

    def generic_visit(self, node: Node, visited_children):
        """Ignore current node, but return children (if any) as a flat list."""
        flat_result = []
        for child in visited_children:
            if type(child) is list:
                flat_result += child  # Unpack child array
            elif child:
                flat_result.append(child)
        return flat_result if len(flat_result) > 0 else None


class RelationParser:
    def __init__(self, query: dict):
        self._query = query
        self._condition_mapping = map_condition_names(query)
        self._used_conditions = None

        found_relations = PathVisitor(self._query, 'query.relation').list()
        assert len(found_relations) in [0, 1]
        self._relation = found_relations[0].strip().lower() if found_relations else None

    def parse(self) -> List[RBaseElement]:
        if not self._relation:
            return []

        ctx = RelationParserContext(condition_count=self._condition_mapping.count,
                                    named_conditions=self._condition_mapping.names,
                                    column_prefix=CONDITION_COLUMN_PREFIX)
        try:
            tree = RELATION_EXPRESSION_GRAMMAR.parse(self._relation)
        except ParseError as pe:
            # Adopted from within the ParseError class, but without the sometimes-confusing issue
            excerpt = pe.text[pe.pos:pe.pos + 20] if (pe.text and pe.pos is not None) else None
            if excerpt:
                message = f"Query relation is invalid around '{excerpt}' "
            else:
                message = f"Query relation '{self._relation}' is invalid"
            raise QueryValidationError(message, kind=ValidationErrorKind.RELATION)

        try:
            elements = RelationExpressionVisitor(ctx).visit(tree)
        except VisitationError as ve:
            logger.exception('Unexpected error while visiting parse tree')
            raise QueryValidationError(message=str(ve), kind=ValidationErrorKind.UNEXPECTED)

        for e in elements:
            error_message = e.validate()
            if error_message:
                raise QueryValidationError(message=error_message, kind=ValidationErrorKind.RELATION)

        self._used_conditions = [e.condition_id for e in elements if e.condition_id is not None]
        return elements

    @property
    def used_conditions(self) -> List[str]:
        return self._used_conditions



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math
import time
from typing import cast
import numpy as np
from frocket.common.config import config
from frocket.common.dataset import DatasetInfo, DatasetId, DatasetPartId, DatasetColumnType
from frocket.common.helpers.utils import timestamped_uuid, bytes_to_ndarray
from frocket.common.tasks.base import TaskStatus
from frocket.common.tasks.registration import RegistrationTaskRequest, RegistrationTaskResult
from frocket.datastore.registered_datastores import get_blobstore, get_datastore
from frocket.worker.runners.registration_task_runner import TOP_GRACE_FACTOR
from tests.utils.task_and_job_utils import simple_run_task
from tests.utils.dataset_utils import str_and_none_column_values, STR_CAT_FEW_WEIGHTS, STR_CAT_MANY_WEIGHTS, \
    TestColumn, DEFAULT_ROW_COUNT, DEFAULT_GROUP_COUNT, DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN, BASE_TIME, \
    TIME_SHIFT, UNSUPPORTED_COLUMN_DTYPES, datafile_schema
from tests.utils.base_test_utils import TEMP_DIR, temp_filename
# noinspection PyUnresolvedReferences
from tests.utils.dataset_utils import datafile
# noinspection PyUnresolvedReferences
from tests.utils.redis_fixture import init_test_redis_settings


def run_task(datafile: str,
             group_id_column: str = DEFAULT_GROUP_COLUMN,
             timestamp_column: str = DEFAULT_TIMESTAMP_COLUMN,
             return_uniques: bool = False) -> RegistrationTaskResult:
    dsid = DatasetId.now(name=timestamped_uuid('ds-'))
    ds = DatasetInfo(id=dsid,
                     basepath=TEMP_DIR,
                     total_parts=1,
                     group_id_column=group_id_column,
                     timestamp_column=timestamp_column)
    req = RegistrationTaskRequest(dataset=ds,
                                  attempt_no=0,
                                  invoke_time=time.time(),
                                  part_id=DatasetPartId(dataset_id=dsid, part_idx=0, path=datafile),
                                  request_id=timestamped_uuid(),
                                  return_group_ids=return_uniques,
                                  invoker_set_task_index=0)

    result = simple_run_task(req, RegistrationTaskResult)
    get_datastore().cleanup_request_data(req.request_id)
    return cast(RegistrationTaskResult, result)


def test_col_types(datafile):
    result = run_task(datafile)
    assert result.status == TaskStatus.ENDED_SUCCESS
    assert result.error_message is None
    sch = result.dataset_schema
    assert sch.group_id_column == DEFAULT_GROUP_COLUMN
    assert sch.timestamp_column == DEFAULT_TIMESTAMP_COLUMN
    assert sch.unsupported_columns == UNSUPPORTED_COLUMN_DTYPES
    assert result.group_ids_blob_id is None

    col = sch.columns[TestColumn.bool]
    assert col.to_dict() == {
        "name": TestColumn.bool, "dtypeName": "bool", "coltype": DatasetColumnType.BOOL,
        "colattrs": {
            "numericMin": None, "numericMax": None, "categorical": False,
            "catTopValues": None, "catUniqueRatio": None}}

    col = sch.columns[TestColumn.float_all_none]
    assert col.name == TestColumn.float_all_none and col.dtype_name == "float64" and \
           col.coltype == DatasetColumnType.FLOAT and np.isnan(col.colattrs.numeric_min) and \
           np.isnan(col.colattrs.numeric_max) and not col.colattrs.categorical

    col = sch.columns[TestColumn.str_object_all_none]
    assert col.coltype == DatasetColumnType.STRING and not col.colattrs.numeric_min and \
           not col.colattrs.numeric_max and not col.colattrs.categorical

    col = sch.columns[TestColumn.str_all_none]
    assert col.coltype == DatasetColumnType.STRING and not col.colattrs.numeric_min and \
           not col.colattrs.numeric_max and not col.colattrs.categorical

    col = sch.columns[TestColumn.int_64_userid]
    assert col.coltype == DatasetColumnType.INT and \
           col.colattrs.numeric_min == 0 and col.colattrs.numeric_max == DEFAULT_GROUP_COUNT - 1 and \
           not col.colattrs.categorical

    col = sch.columns[TestColumn.str_userid]
    assert col.coltype == DatasetColumnType.STRING and not col.colattrs.categorical

    col = sch.columns[TestColumn.str_and_none]
    assert col.coltype == DatasetColumnType.STRING and col.colattrs.categorical
    expected_values = str_and_none_column_values(part=0, with_none=False)  # See datafile creation
    assert col.colattrs.cat_unique_ratio == len(expected_values) / DEFAULT_ROW_COUNT
    assert sorted(col.colattrs.cat_top_values.keys()) == sorted(expected_values)
    assert np.isclose(sum(col.colattrs.cat_top_values.values()), 1.0)

    col = sch.columns[TestColumn.int_64_ts]
    assert col.coltype == DatasetColumnType.INT and \
           col.colattrs.numeric_min == BASE_TIME \
           and col.colattrs.numeric_max == BASE_TIME + TIME_SHIFT

    col = sch.columns[TestColumn.float_64_ts]
    assert col.coltype == DatasetColumnType.FLOAT and \
           math.floor(col.colattrs.numeric_min) == BASE_TIME and \
           math.floor(col.colattrs.numeric_max) == BASE_TIME + TIME_SHIFT

    col = sch.columns[TestColumn.int_u32]
    assert col.coltype == DatasetColumnType.INT

    col = sch.columns[TestColumn.float_32]
    assert col.coltype == DatasetColumnType.FLOAT

    col = sch.columns[TestColumn.str_category_userid]
    assert col.coltype == DatasetColumnType.STRING and col.colattrs.categorical

    col = sch.columns[TestColumn.str_category_few]
    assert col.coltype == DatasetColumnType.STRING and col.colattrs.categorical

    col = sch.columns[TestColumn.str_category_many]
    assert col.coltype == DatasetColumnType.STRING and col.colattrs.categorical

    col = sch.columns[TestColumn.float_category]
    assert col.coltype == DatasetColumnType.FLOAT and not col.colattrs.categorical and \
           math.floor(col.colattrs.numeric_min) == BASE_TIME and \
           math.floor(col.colattrs.numeric_max) == BASE_TIME + TIME_SHIFT

    short_schema_dict = sch.short().to_dict()
    print("Found\n", sch.short().to_json(indent=2))
    print("Expected\n", datafile_schema().to_json(indent=2))
    assert short_schema_dict == datafile_schema().to_dict()


def test_missing_group_column(datafile):
    result = run_task(datafile, group_id_column="mosh")
    assert result.status == TaskStatus.ENDED_FAILED
    result = run_task(datafile, group_id_column="lists")  # Name of an unsupported column
    assert result.status == TaskStatus.ENDED_FAILED


def test_str_group_column_ok(datafile):
    result = run_task(datafile, group_id_column=TestColumn.str_userid)
    assert result.status == TaskStatus.ENDED_SUCCESS


def test_group_column_with_none_notok(datafile):
    result = run_task(datafile, group_id_column=TestColumn.str_and_none)
    assert result.status == TaskStatus.ENDED_FAILED


def test_float_group_column_notok(datafile):
    result = run_task(datafile, group_id_column=TestColumn.float_64_ts)
    assert result.status == TaskStatus.ENDED_FAILED


def test_missing_timestamp_column(datafile):
    result = run_task(datafile, timestamp_column="mosh")
    assert result.status == TaskStatus.ENDED_FAILED


def test_float_timestamp_column_ok(datafile):
    result = run_task(datafile, timestamp_column=TestColumn.float_64_ts)
    assert result.status == TaskStatus.ENDED_SUCCESS


def test_ts_column_with_none_notok(datafile):
    result = run_task(datafile, timestamp_column=TestColumn.float_32)
    assert result.status == TaskStatus.ENDED_FAILED


def test_missing_file():
    result = run_task("sdlasdghjasghdjahsdk")
    assert result.status == TaskStatus.ENDED_FAILED


def test_invalid_file():
    tmpname = temp_filename()
    with open(tmpname, 'w') as f:
        f.write('abc' * 10000)
    result = run_task(tmpname)
    assert result.status == TaskStatus.ENDED_FAILED


def test_uniques_blob(datafile):
    result = run_task(datafile, return_uniques=True)
    assert result.status == TaskStatus.ENDED_SUCCESS
    assert result.group_ids_blob_id
    blob_content = get_blobstore().read_blob(result.group_ids_blob_id)
    assert blob_content
    uniques = bytes_to_ndarray(blob_content)
    assert len(uniques) == DEFAULT_GROUP_COUNT
    assert np.isin(uniques, range(DEFAULT_GROUP_COUNT)).all()

    assert get_blobstore().delete_blob(result.group_ids_blob_id)
    assert not get_blobstore().read_blob(result.group_ids_blob_id)
    assert not get_blobstore().delete_blob(result.group_ids_blob_id)


def test_top_list(datafile):
    result = run_task(datafile)
    assert result.status == TaskStatus.ENDED_SUCCESS
    assert result.dataset_schema.columns[TestColumn.str_category_few].colattrs.cat_top_values == \
           {str(i): w for i, w in enumerate(STR_CAT_FEW_WEIGHTS)}

    long_top_values = result.dataset_schema.columns[TestColumn.str_category_many].colattrs.cat_top_values
    assert long_top_values['0'] == STR_CAT_MANY_WEIGHTS[0]
    assert long_top_values['1'] == STR_CAT_MANY_WEIGHTS[1]
    assert len(long_top_values) == math.floor(config.int('dataset.categorical.top.count') * TOP_GRACE_FACTOR)



"""
implementation of CLI commands.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import json
import sys
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any
from tabulate import tabulate
from frocket.common.config import config
from frocket.common.serializable import SerializableDataClass
from frocket.common.tasks.base import BaseApiResult
from frocket.common.tasks.registration import DatasetValidationMode, RegisterArgs
from frocket.invoker import invoker_api

DATE_FORMAT = '%Y-%m-%d %H:%M:%S %Z'


def run_command(cmd: str, args: argparse.Namespace):
    mapping = {
        'register': register_dataset_cmd,
        'unregister': unregister_dataset_cmd,
        'list': list_datasets_cmd,
        'run': run_query_cmd,
        'info': dataset_info_cmd,
        'config': show_config_cmd
    }
    mapping[cmd](args)


def fail_missing_dataset(name: str):
    sys.exit(f"Dataset '{name}' not found!")


def trim_column(s: str, args: argparse.Namespace, maxwidth: int) -> str:
    if args.notrim or args.nopretty or len(s) <= maxwidth:
        return s
    else:
        return s[:maxwidth - 3] + '...'


def print_json(name: str, o: Any, pretty_print: bool):
    def to_json(o: Any, indent: int = None) -> str:
        return o.to_json(indent=indent) if isinstance(o, SerializableDataClass) else json.dumps(o, indent=indent)

    if pretty_print:
        print(name + ':', to_json(o, indent=2))
    else:
        print(to_json(o))


def handle_api_result(res: BaseApiResult, pretty_print: bool):
    print_json('API Result', res, pretty_print)
    if not res.success:
        sys.exit('FAILED' if pretty_print else 1)


def register_dataset_cmd(args):
    validation_mode = DatasetValidationMode[args.validation.upper()]
    register_args = RegisterArgs(name=args.name,
                                 basepath=args.basepath,
                                 group_id_column=args.group_id_column,
                                 timestamp_column=args.timestamp_column,
                                 pattern=args.pattern,
                                 validation_mode=validation_mode,
                                 validate_uniques=not args.skip_uniques)
    res = invoker_api.register_dataset(register_args)
    handle_api_result(res, pretty_print=not args.nopretty)


def unregister_dataset_cmd(args):
    res = invoker_api.unregister_dataset(args.dataset, force=args.force)
    handle_api_result(res, pretty_print=not args.nopretty)


def list_datasets_cmd(args):
    datasets = sorted(invoker_api.list_datasets(), key=lambda ds: ds.id.registered_at, reverse=True)
    display_datasets = [{'name': trim_column(ds.id.name, args, maxwidth=30),
                         'registered at': ds.id.registered_at.strftime(DATE_FORMAT),
                         'parts': ds.total_parts,
                         'group id': ds.group_id_column,
                         'timestamp': ds.timestamp_column,
                         'path': trim_column(ds.basepath, args, maxwidth=50)}
                        for ds in datasets]
    if args.nopretty:
        print(json.dumps(display_datasets))
    else:
        if len(datasets) == 0:
            print('No datasets registered yet')
        else:
            print(tabulate(display_datasets, headers='keys'))


def json_parse(s: str) -> dict:
    try:
        return json.loads(s)
    except JSONDecodeError as e:
        sys.exit(f'JSON Error: {e}')


def run_query_cmd(args):
    ds_info = invoker_api.get_dataset(args.dataset)
    if not ds_info:
        fail_missing_dataset(args.dataset)
    query = None
    if args.empty:
        query = {}
    elif args.query_string:
        query = json_parse(args.query_string)
    elif args.filename:
        filepath = Path(args.filename)
        if not filepath.exists():
            sys.exit(f'File not found: {args.filename}')
        else:
            query_str = filepath.read_text(encoding='utf-8')
            query = json_parse(query_str)
    else:
        sys.exit('Unknown mode')

    try:
        res = invoker_api.run_query(ds_info, query)
        handle_api_result(res, pretty_print=not args.nopretty)
    except Exception as e:
        sys.exit(f'Error: {e}')


def dataset_info_cmd(args):
    show_full = args.full
    ds_info = invoker_api.get_dataset(args.dataset)
    if not ds_info:
        fail_missing_dataset(args.dataset)
    parts_info = invoker_api.get_dataset_parts(ds_info)
    schema_info = invoker_api.get_dataset_schema(ds_info, full=show_full)
    print_json('Basic information', ds_info, pretty_print=not args.nopretty)
    print_json('Parts', parts_info, pretty_print=not args.nopretty)
    print_json(f'Schema (full: {show_full})', schema_info, pretty_print=not args.nopretty)


def show_config_cmd(args):
    print_json(f'Configuration', config, pretty_print=not args.nopretty)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, Type, Callable, List, cast
from frocket.common.validation.relation_parser import RBaseElement, RTextElement, RConditionBaseElement, ROperator


def relation_to_pandas_query(elements: List[RBaseElement], column_prefix: str) -> str:
    """Convert the generic pasred representation of query.relation expression (as returned by QueryValidator or its
    helper class RelationParser) into a Pandas query string."""

    # Mapping of generic element type to a lambda function constructing the Pandas equivalent. Note below that not
    # every concreate element type needs an entry here, as the code would also look for its superclass
    etype_to_handler: Dict[Type[RBaseElement], Callable[[RBaseElement], str]] = {
        RTextElement: lambda v: v.text,
        RConditionBaseElement: lambda v: f"{column_prefix}{v.condition_id}",
        ROperator: lambda v: " & " if v.text in ["and", "&&"] else " | "
    }

    transformed = []
    for e in elements:
        func = None
        # Either there's a handler above for this element type, or go up the superclass chain to find one.
        class_and_supers = cast(List[Type[RBaseElement]], type(e).mro())
        for cls in class_and_supers:
            func = etype_to_handler.get(cls, None)
            if func:
                break
        if not func:
            raise Exception(f"{e} has no handler for any of its superclasses: {class_and_supers}")
        transformed.append(func(e))
    return "".join(transformed)
